"""
Created on May 26, 2015

@author: Derek Wood
"""
# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import functools
import os
import threading
from collections import defaultdict
from operator import attrgetter
from typing import *

from sqlalchemy.sql.schema import Column

from bi_etl.components.row.cached_frozenset import get_cached_frozen_set

if TYPE_CHECKING:
    import bi_etl.components.row.row.Row
    from bi_etl.components.etlcomponent import ETLComponent


class RowIterationHeader(object):
    """
    Stores the headers of a set of rows for a given iteration
    """
    lock = threading.Lock()
    next_iteration_id = 0
    instance_dict = dict()
    __shared_name_map_db = dict()
    _request_header_info_thread = None
    _request_header_info_queue = None
    _process_instance_dict = defaultdict(dict)

    @staticmethod
    def get_by_id(iteration_id):
        return RowIterationHeader.instance_dict[iteration_id]

    @staticmethod
    def add_remote_iteration_header(remote_header):
        RowIterationHeader._process_instance_dict[remote_header.owner_pid][remote_header.iteration_id] = remote_header

    @staticmethod
    def get_by_process_and_id(value_sent):
        # value_sent needs to match RowIterationHeader.get_cross_process_iteration_header return
        process_pid, process_iteration_id = value_sent
        if process_pid == os.getpid():
            return RowIterationHeader.get_by_id(process_iteration_id)
        elif process_pid in RowIterationHeader._process_instance_dict:
            if process_iteration_id in RowIterationHeader._process_instance_dict[process_pid]:
                return RowIterationHeader._process_instance_dict[process_pid][process_iteration_id]
        # Else not found
        raise ValueError(
            f"Iteration header for PID={process_pid} ID={process_iteration_id} has not been sent to this process"
        )

    def __init__(
            self,
            logical_name: Optional[str] = None,
            primary_key: Optional[Iterable] = None,
            parent: Optional[ETLComponent] = None,
            columns_in_order: Optional[Iterable] = None,
            owner_pid: int = None,
            ):
        with RowIterationHeader.lock:
            RowIterationHeader.next_iteration_id += 1
            self.iteration_id = RowIterationHeader.next_iteration_id
            RowIterationHeader.instance_dict[self.iteration_id] = self
        self.column_definition_locked = False
        if columns_in_order:
            columns_in_order = list(columns_in_order)
            self._column_count = len(columns_in_order)
        else:
            self._column_count = 0
        self.owner_pid = owner_pid
        self.row_count = 0
        self.logical_name = logical_name or id(self)
        self._primary_key = None
        self.primary_key = primary_key
        self.parent = parent
        self._actions_to_next_headers = dict()
        self._columns_positions = dict()
        self._name_map_db = self.__shared_name_map_db
        if columns_in_order is not None:
            self._columns_in_order = None
            self.columns_in_order = columns_in_order
            with RowIterationHeader.lock:
                self._name_map_db.update({col: col for col in columns_in_order})
        else:
            self._columns_in_order = list()
        self.columns_frozen = False
        self._cached_column_set = None
        self._cached_positioned_column_set = None
        self._cached_columns_in_order_tuple = None
        self._action_position = None
        self.action_id = None
        self.action_count = 0

    def get_cross_process_iteration_header(self):
        return (
            os.getpid(),
            self.iteration_id,
        )

    def __reduce__(self):
        return (
            # A callable object that will be called to create the initial version of the object.
            self.__class__,

            # A tuple of arguments for the callable object. An empty tuple must be given if the callable does not accept any argument
            (
                self.logical_name,
                self.primary_key,
                self.parent,
                self.columns_in_order,
                os.getpid(),
            ),
            # Optionally, the object’s state, which will be passed to the object’s __setstate__() method as previously described.
            # If the object has no such method then, the value must be a dictionary and it will be added to the object’s __dict__ attribute.

            # Optionally, an iterator (and not a sequence) yielding successive items.
            # These items will be appended to the object either using obj.append(item) or, in batch, using obj.extend(list_of_items).

            # Optionally, an iterator (not a sequence) yielding successive key-value pairs.
            # These items will be stored to the object using obj[key] = value

            # PROTOCOL 5+ only
            # Optionally, a callable with a (obj, state) signature.
            # This callable allows the user to programmatically control the state-updating behavior of a specific object,
            # instead of using obj’s static __setstate__() method.
            # If not None, this callable will have priority over obj’s __setstate__().
        )

    def get_column_name(self, input_name: Union[str, Column]) -> str:
        try:
            with RowIterationHeader.lock:
                return self._name_map_db[input_name]
        except KeyError:
            # If the input_name is an SA Column use it's name.
            # In Python 2.7 to 3.4, isinstance is a lot faster than try-except or hasattr (which does a try)
            if isinstance(input_name, str):
                name_str = input_name
            elif isinstance(input_name, Column):
                name_str = input_name.name
            else:
                raise ValueError(f"Row column name must be str, unicode, or Column. Got {type(input_name)}")
            with RowIterationHeader.lock:
                self._name_map_db[input_name] = name_str
            return name_str

    def get_action_header(
            self,
            action: tuple,
            start_empty: bool = False,
            ) -> RowIterationHeader:
        """
        Get the header after performing a manipulation on the set of columns.

        Parameters
        ----------
        action:
            A hashable action ID
        start_empty:
            Should the new header start empty (vs transferring the columns)

        Returns
        -------

        """
        return self._actions_to_next_headers[action]

    def get_next_header(
            self,
            action: tuple,
            start_empty: bool = False,
            ) -> RowIterationHeader:
        """
        Get the next header after performing a manipulation on the set of columns.

        Parameters
        ----------
        action:
            A hashable action ID
        start_empty:
            Should the new header start empty (vs transferring the columns)

        Returns
        -------

        """
        if action not in self._actions_to_next_headers:
            new_header = self.__class__(
                logical_name=self.logical_name,
                primary_key=self.primary_key,
                parent=None
                )
            new_header.action_count = self.action_count + 1
            new_header.action_id = action
            if not start_empty:
                new_header._columns_in_order = self._columns_in_order.copy()
                new_header._columns_positions = self._columns_positions.copy()
                new_header._column_count = self._column_count

            self._actions_to_next_headers[action] = new_header
        return self._actions_to_next_headers[action]

    @functools.lru_cache()
    def __repr__(self):
        return f'{self.__class__.__name__}' \
               f'(id={self.iteration_id},logical_name={self.logical_name},primary_key={self.primary_key}'

    @functools.lru_cache()
    def __str__(self):
        return repr(self)

    @property
    def columns_in_order(self) -> Sequence:
        """
        A list of the columns of this row in the order they were defined.
        """
        if self._cached_columns_in_order_tuple is None:
            self._cached_columns_in_order_tuple = tuple(self._columns_in_order)
        return self._cached_columns_in_order_tuple

    @columns_in_order.setter
    def columns_in_order(self, value: Iterable):
        if self._columns_in_order is not None:
            if self.column_count > 0:
                raise ValueError("Setting columns_in_order is only allowed on an empty RowIterationHeader")
        self._columns_in_order = list(value)
        for position, column_name in enumerate(value):
            self._columns_positions[column_name] = position
        self._clear_caches()

    @property
    def primary_key(self) -> Optional[List]:
        try:
            if self._primary_key is not None and len(self._primary_key) > 0:
                # Check if primary_key is list of Column objects and needs to be turned into a list of str name values
                # We do this here and not on setter since program might never call getter
                if isinstance(self._primary_key[0], Column):
                    self._primary_key = list(map(attrgetter('name'), self._primary_key))
                return self._primary_key
            else:
                return None
        except AttributeError:
            return None

    @primary_key.setter
    def primary_key(self, value: Optional[Union[str, Iterable]]):
        if value is not None:
            if isinstance(value, str):
                value = [value]
            elif not isinstance(value, list):
                value = list(value)
            assert hasattr(value, '__iter__'), "primary_key must be iterable or string"
            self._primary_key = value
        else:
            self._primary_key = None

    @property
    def column_set(self) -> frozenset:
        """
        An ImmutableSet of the columns of this row.
        Used to store different row configurations in a dictionary or set.

        WARNING: The resulting set is not ordered. Do not use if the column order affects the operation.
        See positioned_column_set instead.
        """
        if self._cached_column_set is None:
            self._cached_column_set = frozenset(self._columns_in_order)
        return self._cached_column_set

    @property
    def positioned_column_set(self) -> Set[tuple]:
        """
        An ImmutableSet of the tuples (column, position) for this row.
        Used to store different row configurations in a dictionary or set.

        Note: column_set would not always work here because the set is not ordered even though the columns are.

        """
        if self._cached_positioned_column_set is None:
            tpl_lst = list()
            for key, position in enumerate(self._columns_in_order):
                tpl = (key, position)
                tpl_lst.append(tpl)
            self._cached_positioned_column_set = get_cached_frozen_set(tpl_lst)
        return self._cached_positioned_column_set

    def add_row(self, row):
        self.row_count += 1

    def remove_row(self, row):
        self.row_count -= 1

    @property
    def column_count(self) -> int:
        return self._column_count

    def has_column(self, column_name) -> bool:
        return column_name in self._columns_positions

    def _key_error(self, column_name) -> KeyError:
        return KeyError(
            f"{self.logical_name} has no item {column_name} "
            f"it does have {self._columns_in_order}"
            )

    @functools.lru_cache(maxsize=1000)
    def get_column_position(
            self,
            column_name: str,
            allow_create: bool = False,
            ) -> int:
        """
        Get the ordinal column position based on a column name (str)

        Parameters
        ----------

        column_name:
            String column name
        allow_create:
            Is this method allowed to create a new column.
            Note: if :py:attr:`columns_frozen` is True this method will return a KeyError
            even if allow_create is True.
        """
        try:
            position = self._columns_positions[column_name]
        except KeyError:
            if self.columns_frozen or not allow_create:
                raise self._key_error(column_name)
            else:
                position = self._add_column(column_name)
        return position

    # noinspection PyProtectedMember
    def row_set_item(
            self,
            column_name: str,
            value,
            row: bi_etl.components.row.row.Row,
            ) -> RowIterationHeader:
        """
        Set a column in a row and return a new row header (it might have changed if the column was new). 
        
        Parameters:        
            column_name: column to set
            value: new value
            row (bi_etl.components.row.row.Row): row to find column on

        Returns:        
            Modified row header
        """
        if column_name in self._columns_positions:
            position = self._columns_positions[column_name]
            # noinspection PyProtectedMember
            row.set_by_zposition(position, value)
            new_header = self
        else:
            # Modification of columns required
            action_tuple = ('+:', column_name)
            new_header = self.get_next_header(action_tuple)
            new_header.add_row(row)
            if new_header._action_position is None:
                new_header._action_position = new_header._add_column(column_name)
            else:
                pass

            # Protected access is required here since we can't call setitem it calls this method.
            # noinspection PyProtectedMember
            row._data_values.append(value)
        return new_header

    def rename_column(
            self,
            old_name: str,
            new_name: str,
            ignore_missing: bool = False,
            no_new_header: bool = False,
            ) -> RowIterationHeader:
        """
        Rename a column

        Parameters:
            old_name: str
                The name of the column to find and rename.
    
            new_name: str
                The new name to give the column.
    
            ignore_missing: boolean
                Ignore (don't raise error) if we don't have a column with the name in old_name.
                Defaults to False
    
            no_new_header:
                Skip creating a new row header, modify in place.
                
                ** BE CAREFUL USING THIS! **
                
                All new rows created with this header will immediately get the new name,
                in which case you won't want to call this method again.
        """
        # Clear LRU cache of get_column_position
        self.get_column_position.cache_clear()
        try:
            position = self._columns_positions[old_name]
            # Only check if target column name exists if we found the source name
            # This way we can have two renames that both map to the same target as long as
            # 1) Only one of the source column names exists
            # 2) ignore_missing is True
            assert new_name not in self._columns_positions, f"Target column name {new_name} already exists"
            # Modification of columns required
            if not no_new_header:
                action_tuple = ('r', old_name, new_name)
                new_header = self.get_next_header(action_tuple)
                new_header.row_count = self.row_count
            else:
                new_header = self
            if old_name in new_header._columns_positions:
                new_header._columns_in_order[position] = new_name
                del new_header._columns_positions[old_name]
                new_header._columns_positions[new_name] = position
                if new_header.primary_key is not None:
                    try:
                        pk_position = new_header.primary_key.index(old_name)
                        new_header.primary_key[pk_position] = new_name
                    except ValueError:
                        pass
        except (KeyError, ValueError) as e:
            if ignore_missing:
                return self
            else:
                raise ValueError(
                    f'Rename error: {old_name} is not a column in this Row. '
                    f'Valid columns in {self.logical_name} from {self.parent} are {self._columns_positions.keys()}'
                )
        return new_header

    def rename_columns(
            self,
            rename_map: Union[dict, List[tuple]],
            ignore_missing: bool = False,
            no_new_header: bool = False
            ) -> RowIterationHeader:
        """
        Rename many columns at once.

        Parameters:
            rename_map
                A dict or list of tuples to use to rename columns.
                Note a list of tuples is better to use if the renames need to happen in a certain order.
    
            ignore_missing:
                Ignore (don't raise error) if we don't have a column with the name in old_name.
                Defaults to False
    
            no_new_header:
                Skip creating a new row header, modify in place.
                
                ** BE CAREFUL USING THIS! **
                
                All new rows created with this header will immediately get the new name,
                in which case you won't want to call this method again.
        """
        new_header = self
        # Note: By using id(rename_map) we assume that the calling code will not change
        # the rename map after we first see it. This seems like a reasonable assumption
        # for the performance gained.
        action_tuple = ('rc', id(rename_map))
        try:
            new_header = self._actions_to_next_headers[action_tuple]
        except KeyError:
            # Results of rename not already stored, perform the renames one by one
            if isinstance(rename_map, Mapping):
                for k in rename_map.keys():
                    new_header = new_header.rename_column(k, rename_map[k],
                                                          ignore_missing=ignore_missing,
                                                          no_new_header=no_new_header)
            elif rename_map is not None:  # assume it's a list of tuples
                for (old, new) in rename_map:
                    new_header = new_header.rename_column(old, new,
                                                          ignore_missing=ignore_missing,
                                                          no_new_header=no_new_header)
            # Now store the shortcut
            self._actions_to_next_headers[action_tuple] = new_header

        return new_header

    def row_remove_column(
            self,
            column_name: str,
            row: bi_etl.components.row.row.Row,
            ignore_missing: bool = False,
            ) -> RowIterationHeader:
        if column_name not in self._columns_positions:
            if not ignore_missing:
                raise self._key_error(column_name)
            return self
        else:
            # Modification of columns required
            self.get_column_position.cache_clear()
            action_tuple = ('-:', column_name)
            new_header = self.get_next_header(action_tuple)
            new_header.add_row(row)
            if new_header._action_position is None:
                position = new_header._columns_positions[column_name]
                new_header._action_position = position
                del new_header._columns_in_order[position]
                del new_header._columns_positions[column_name]
                for following_col in new_header._columns_in_order[position:]:
                    new_header._columns_positions[following_col] -= 1
                new_header._column_count = len(new_header._columns_in_order)
                if self.primary_key is not None and column_name in self.primary_key:
                    self.primary_key.remove(column_name)
            # Protected access is required here since we can't call __delitem__, it calls this method.
            # noinspection PyProtectedMember
            del row._data_values[new_header._action_position]
            return new_header

    def row_subset(
            self,
            row: bi_etl.components.row.row.Row,
            exclude: Optional[Iterable] = None,
            rename_map: Optional[Union[dict, List[tuple]]] = None,
            keep_only: Optional[Iterable] = None,
            ) -> bi_etl.components.row.row.Row:
        """
        Return a new row instance with a subset of the columns. Original row is not modified
        Excludes are done first, then renames and finally keep_only.

        Parameters
        ----------
        row:
            The row to subset
        exclude:
            A list of column names (before renames) to exclude from the subset.
            Optional. Defaults to no excludes.

        rename_map:
            A dict to use to rename columns.
            Optional. Defaults to no renames.

        keep_only:
            A list of column names (after renames) of columns to keep.
            Optional. Defaults to keep all.

        Returns
        -------
        a list with the position mapping of new to old items.
        So:
            The first item in the list will be the index of that item in the old list.
            The second item in the list will be the index of that item in the old list.
            etc
        """
        action_1 = 's'
        action_2 = frozenset(exclude or {})
        action_3 = frozenset(keep_only or {})
        # Note: By using id(rename_map) we assume that the calling code will not change
        # the rename map after we first see it. This seems like a reasonable assumption
        # for the performance gained.
        action_4 = id(rename_map)

        action_tuple = (action_1, action_2, action_3, action_4)

        new_iteration_header = self.get_next_header(action_tuple)
        sub_row = row.__class__(iteration_header=new_iteration_header, allocate_space=False)

        if new_iteration_header._action_position is None:
            # Do the transformation for the first time
            if rename_map is not None:
                new_iteration_header.rename_columns(rename_map, no_new_header=True)
            old_positions_list = list()
            sub_pos = 0
            for parent_position, old_column_name in enumerate(self._columns_in_order):
                remove_column = True
                new_column_name = new_iteration_header._columns_in_order[sub_pos]
                if old_column_name not in exclude:
                    if keep_only is None or new_column_name in keep_only:
                        old_positions_list.append(parent_position)
                        new_iteration_header._columns_positions[new_column_name] = sub_pos
                        sub_pos += 1
                        remove_column = False
                if remove_column:
                    del new_iteration_header._columns_in_order[sub_pos]
                    del new_iteration_header._columns_positions[new_column_name]
            new_iteration_header._column_count = len(new_iteration_header._columns_in_order)
            new_iteration_header._action_position = old_positions_list
            # These should already be None unless running in the debugger, since it might have called
            # on the column_set or positioned_column_set properties.
            new_iteration_header._clear_caches()
        # Build the new row based on the _action_position details
        # Protected access is required here since we can't call setitem it calls this method.
        # noinspection PyProtectedMember
        sub_row._data_values = [row._data_values[old_pos] for old_pos in new_iteration_header._action_position]
        return sub_row

    def _add_column(
            self,
            column_name: str,
            ) -> int:
        position = self._column_count
        self._column_count += 1
        self._columns_in_order.append(column_name)
        self._columns_positions[column_name] = position
        self._clear_caches()
        return position

    def _clear_caches(self):
        self._cached_column_set = None
        self._cached_positioned_column_set = None
        self._cached_columns_in_order_tuple = None



import logging


class NotifierBase(object):
    def __init__(self):
        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")

    def send(self, subject, message, sensitive_message=None, attachment=None, throw_exception=False):
        pass

    def post_status(self, status_message):
        """
        Send a temporary status messages that gets overwritten with the next status message that is sent.

        Parameters
        ----------
        status_message

        Returns
        -------

        """
        raise NotImplementedError("This Notifier does not implement post_status")


class NotifierException(Exception):
    pass



from bi_etl.components.get_next_key.base_table_memory import BaseTableMemory


class LocalTableMemory(BaseTableMemory):
    def __init__(self, table):
        super().__init__(table=table)
        self.current_key_values = dict()

    def get_next_key(self, column: str) -> int:
        column_obj = self.table.get_column(column)
        column = column_obj.name
        current_max = self.current_key_values.get(column)
        if current_max is None:
            next_key = self.get_next_from_database(column_obj)
        else:
            next_key = current_max + 1
        self.current_key_values[column] = next_key
        return next_key



"""
Created on Sept 12 2016

@author: Derek
"""
from bi_etl.components.readonlytable import ReadOnlyTable
from bi_etl.database.database_metadata import DatabaseMetadata
from bi_etl.scheduler.task import ETLTask


class CopyTableData(ETLTask):
    def depends_on(self):
        return []

    def load(self):
        database = self.get_parameter('database')
        assert isinstance(database, DatabaseMetadata)
        source_table_name = self.get_parameter('source_table')
        target_table_name = self.get_parameter('target_table')

        with ReadOnlyTable(
                self,
                database,
                source_table_name,
                table_name_case_sensitive=None,
                ) as source_data:
            with ReadOnlyTable(
                    self,
                    database,
                    target_table_name,
                    table_name_case_sensitive=None,
                    ) as target_tbl:
                target_column_set = set(target_tbl.column_names)
                common_columns = list()
                for source_col in source_data.column_names:
                    if source_col in target_column_set:
                        common_columns.append(source_col)

                cols = ""
                sep = ""
                for column in common_columns:
                    cols += sep
                    cols += column
                    sep = ","

                sql = "INSERT INTO {target_table_name} ({cols})"\
                      "SELECT {cols} FROM {source_table_name}"\
                      .format(
                         source_table_name=source_table_name,
                         target_table_name=target_table_name,
                         cols=cols)

                self.log.debug(sql)

                database.execute(sql, transaction=True)

        self.log.info("Done")



from typing import List

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_templates.credentials import Credentials
from config_wrangler.config_templates.keepass_config import KeepassConfig
from config_wrangler.config_templates.password_source import PasswordSource


class NotifierConfigBase(ConfigHierarchy):
    notifier_class: str


class LogNotifierConfig(NotifierConfigBase):
    notifier_class: str = 'bi_etl.notifiers.log_notifier.LogNotifier'


class SMTP_Notifier(NotifierConfigBase, Credentials):
    notifier_class: str = 'bi_etl.notifiers.email.Email'
    email_from: str
    gateway_host: str = None
    gateway_port: int = 0
    use_ssl: bool = False
    debug: bool = False
    distro_list: List[str]


class SlackNotifier(NotifierConfigBase):
    notifier_class: str = 'bi_etl.notifiers.slack.Slack'
    channel: str
    token: str = None
    """
    This is only used for the extremely non-secure `CONFIG_FILE` token source 
    valid values defined using `PasswordSource`.
    The token is stored directly in the config file with the setting 
    name `token`
    """

    mention: str = None

    token_source: PasswordSource = None
    """
    The source to use when getting a token for slack.  
    See :py:class:`PasswordSource` for valid values.
    """

    keyring_section: str = None
    """
    If the password_source is KEYRING, then which section (AKA system)
    should this module look for the password in.
    
    See https://pypi.org/project/keyring/
    or https://github.com/jaraco/keyring
    """
    keepass_config: str = 'keepass'
    """
    If the password_source is KEEPASS, then which root level config item contains
    the settings for Keepass (must be an instance of 
    :py:class:`config_wrangler.config_templates.keepass_config.KeepassConfig`)
    """
    keepass: KeepassConfig = None
    """
    If the password_source is KEEPASS, then load a sub-section with the 
    :py:class:`config_wrangler.config_templates.keepass_config.KeepassConfig`) settings
    """

    keepass_group: str = None
    """
    If the password_source is KEEPASS, which group in the Keepass database should
    be searched for an entry with a matching entry.
    
    If is None, then the `KeepassConfig.default_group` value will be checked.
    If that is also None, then a ValueError will be raised.
    """

    keepass_title: str = None
    """
    If the password_source is KEEPASS, this is an optional filter on the title
    of the keepass entries in the group.
    """


class JiraNotifier(NotifierConfigBase, Credentials):
    notifier_class: str = 'bi_etl.notifiers.jira.Jira'
    server: str
    project: str
    component: str = None
    comment_on_each_instance: bool
    exclude_statuses: List[str] = ['Closed']
    issue_type: str = 'Bug'
    priority: str = None
    subject_prefix: str = ''
    comment_on_each_instance: bool = True



import logging


def modify_log_format(log, replace, search='%(message)s'):
    for handler in log.handlers:
        format_str = handler.formatter._fmt
        format_str = format_str.replace(
            search,
            replace
        )
        fmt = logging.Formatter(format_str)
        handler.setFormatter(fmt)
    if log.parent is not None:
        modify_log_format(
            log=log.parent,
            replace=replace,
            search=search
        )


from bi_etl.components.etlcomponent import ETLComponent
from bi_etl.scheduler.task import ETLTask


class StaticReader(ETLComponent):
    ROWS = 10**6
    COLS = 100
    TEXT_COLS = 5
    INT_COLS = COLS - TEXT_COLS

    def _obtain_column_names(self):
        self._column_names = [f'x{i}' for i in range(self.COLS)]

    def _raw_rows(self):
        this_iteration_header = self.full_iteration_header
        for row_num in range(self.ROWS):
            row_list = [f'x{i}[{row_num}]' for i in range(self.TEXT_COLS)]
            row_list.extend([i + row_num for i in range(self.INT_COLS)])
            d = self.Row(data=row_list, iteration_header=this_iteration_header)
            yield d


class TestStaticReader(ETLTask):
    def depends_on(self):
        return []

    def load(self):
        budget_amount = 0
        with StaticReader(self) as source_data:
            for row in source_data:
                budget_amount += row['x6']

        self.log.info(budget_amount)

        self.log.info("Done")


if __name__ == '__main__':
    TestStaticReader().run()




import multiprocessing

from bi_etl.components.get_next_key.shared_table_memory import SharedTableMemory


def main():
    pass


if __name__ == '__main__':
    main()


class SharedTableMemoryManager(object):
    def __init__(self, manager):
        self.table_dict = manager.dict()

    def create_shared_table_memory(self, manager: multiprocessing.Manager, table_name: str, chunk_size: int):
        table_name = str(table_name)
        if table_name not in self.table_dict:
            self.table_dict[table_name] = SharedTableMemory(manager, chunk_size=chunk_size)
        return self.table_dict[table_name]

    def get_shared_table_memory(self, table_name):
        return self.table_dict[table_name]



from pathlib import Path

import toml

package_root = Path(__file__).parent.parent.absolute()
poetry_config = toml.load(str(package_root / 'pyproject.toml'))

full_version = poetry_config['tool']['poetry']['version']

version_parts = full_version.split('.')

version_1 = '.'.join(version_parts[:1])
version_1_2 = '.'.join(version_parts[:2])



# -*- coding: utf-8 -*-
"""
Created on Dec 23, 2015

@author: Derek Wood
"""

import sys

from bi_etl.scheduler.sdtout_queue import StdoutQueue


def redirect_output_to(queue_instance):
    if queue_instance is not None and not hasattr(sys, 'saved_stds_tuple'):
        sys.saved_stds_tuple = (sys.stdout, sys.stderr)
        queue_out_stream = StdoutQueue(queue_instance)
        sys.stdout = queue_out_stream
        sys.stderr = queue_out_stream
        return queue_out_stream
    else:
        return None


def restore_standard_output():
    if hasattr(sys, 'saved_stds_tuple'):
        sys.stdout, sys.stderr = sys.saved_stds_tuple
        del sys.saved_stds_tuple



# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import os.path
import os.path
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from tempfile import TemporaryDirectory
from typing import *

import pyarrow as pa
import pyarrow.parquet as pq

from bi_etl.bulk_loaders.redshift_s3_base import RedShiftS3Base
from bi_etl.bulk_loaders.s3_bulk_load_config import S3_Bulk_Loader_Config

if TYPE_CHECKING:
    from bi_etl.components.table import Table
    from bi_etl.scheduler.task import ETLTask


class RedShiftS3ParquetBulk(RedShiftS3Base):
    def __init__(
        self,
        config: S3_Bulk_Loader_Config,
    ):
        super().__init__(
            config=config,
        )

        if self.s3_file_max_rows is None and self.s3_files_to_generate is None:
            self.s3_file_max_rows = 2_000_000

        self.lines_scanned_modifier = 0

    @property
    def needs_all_columns(self):
        return False

    def get_copy_sql(
            self,
            s3_source_path: str,
            table_to_load: str,
            file_compression: str = '',
            analyze_compression: str = None,
            options: str = '',
    ):
        analyze_compression = analyze_compression or self.analyze_compression
        if analyze_compression:
            options += f' COMPUPDATE {self.analyze_compression} '

        return f"""\
                COPY {table_to_load} FROM 's3://{self.s3_bucket_name}/{s3_source_path}'                      
                     credentials 'aws_access_key_id={self.s3_user_id};aws_secret_access_key={self.s3_password}'
                     FORMAT PARQUET
                     {file_compression}  
                     {options}
               """

    def _write_to_file(
            self,
            temp_dir,
            file_number: int,
            schema: pa.Schema,
            data_chunk: list,
            local_files_list: list,
    ):
        self.log.debug(f"Starting bulk load chunk {file_number}")
        # TODO: Redshift will recognize and uncompress these file types
        #   .gz
        #   .snappy
        #   .bz2
        filepath = os.path.join(temp_dir, f'data_{file_number}.parquet.snappy')
        local_files_list.append(filepath)
        table = pa.Table.from_pylist(data_chunk, schema=schema)
        pq.write_table(table, filepath, compression='SNAPPY')
        self.log.debug(f"Closed bulk load chunk {file_number}: {filepath}")

    def load_from_iterator(
           self,
           iterator: Iterator,
           table_object: Table,
           table_to_load: str = None,
           perform_rename: bool = False,
           progress_frequency: int = 10,
           analyze_compression: str = None,
           parent_task: Optional[ETLTask] = None,
    ) -> int:

        table_to_load = table_to_load or table_object.qualified_table_name

        with TemporaryDirectory(ignore_cleanup_errors=True) as temp_dir:
            # Generate schema from table
            fields = list()
            type_map = {
                str: pa.string(),
                bytes: pa.large_binary(),
                date: pa.date32(),
                datetime: pa.date64(),
                time: pa.time64('us'),
                timedelta: pa.month_day_nano_interval(),
                bool: pa.bool_(),
                int: pa.int64(),
                float: pa.float64(),
            }
            for column in table_object.columns:
                if column.type.python_type in type_map:
                    col_type = type_map[column.type.python_type]
                else:
                    if column.type.python_type == Decimal:
                        col_type = pa.decimal128(
                            column.type.precision,
                            column.type.scale,
                        )
                    else:
                        col_type = pa.string()

                fields.append(
                    pa.field(
                        column.name,
                        col_type,
                        True,  # Nullable
                    )
                )

            schema = pa.schema(fields)

            self.log.debug(f"schema = {schema}")

            file_number = 0
            data_length = 0
            local_files_list = list()

            # Note: Tried using this method to chunk the list, but it didn't work with the
            # gevent based threading used by insert_row with bulk loader
            # for data_chunk in more_itertools.chunked(iterator, self.s3_file_max_rows):

            data_chunk = list()
            for row in iterator:
                data_chunk.append(row)

                if len(data_chunk) >= self.s3_file_max_rows:
                    data_length += len(data_chunk)
                    file_number += 1
                    self._write_to_file(
                        temp_dir=temp_dir,
                        file_number=file_number,
                        schema=schema,
                        data_chunk=data_chunk,
                        local_files_list=local_files_list,
                    )
                    # TODO: We could start uploading the files to S3 with a gevent "thread" here
                    data_chunk = list()

            # Write final chunk
            if len(data_chunk) >= 0:
                data_length += len(data_chunk)
                file_number += 1
                self._write_to_file(
                    temp_dir=temp_dir,
                    file_number=file_number,
                    schema=schema,
                    data_chunk=data_chunk,
                    local_files_list=local_files_list,
                )

            self.log.debug(f"Loading from files {local_files_list}")
            rows_loaded = self.load_from_files(
                local_files_list,
                table_object=table_object,
                table_to_load=table_to_load,
                perform_rename=perform_rename,
                analyze_compression=analyze_compression,
            )
            if rows_loaded != data_length:
                self.log.error(f"COPY from files should have loaded {data_length:,} but it reports {rows_loaded:,} rows loaded")
            else:
                self.log.info(f"{self} had nothing to do with 0 rows found")
            return data_length



class BulkLoaderException(Exception):
    def __init__(self, base_exception, message=None, password=None):
        self.base_exception = base_exception
        self.message = str(base_exception)
        if message is not None:
            self.message += message
        if password is not None:
            self.remove_password(password)
        self.errors_set = set()

    def add_error(self, error):
        self.errors_set.add(str(error))

    def remove_password(self, password):
        if password in self.message:
            self.message = self.message.replace(password, '*' * 8)

    def __repr__(self):
        return f"BulkLoaderException({type(self.base_exception)},message=({str(self)})"

    def __str__(self):
        msg = f"{self.message})"
        if len(self.errors_set) > 0:
            msg += '\n'
            msg += 'BULK LOAD ERROR SUMMARY:\n'
            msg += '------------------------\n'
            msg += '\n'.join(self.errors_set)
        return msg



import functools
from typing import *

from sqlalchemy import Column

from bi_etl.components.row.row_iteration_header import RowIterationHeader


if TYPE_CHECKING:
    from bi_etl.components.row.row import Row
    from bi_etl.components.etlcomponent import ETLComponent


class RowIterationHeaderCaseInsensitive(RowIterationHeader):
    # Override since the case-sensitive versions should not share a name map
    __shared_name_map_db = dict()

    def __init__(
            self,
            logical_name: Optional[str] = None,
            primary_key: Optional[Iterable] = None,
            parent: Optional['ETLComponent'] = None,
            columns_in_order: Optional[Iterable] = None,
            owner_pid: int = None,
    ):
        if columns_in_order is not None:
            columns_in_order = [c.lower() for c in columns_in_order]
        super().__init__(
            logical_name=logical_name,            
            primary_key=primary_key,            
            parent=parent,            
            columns_in_order=columns_in_order,            
            owner_pid=owner_pid,            
        )

    @classmethod
    def from_other_header(cls, other_header: RowIterationHeader):
        columns_in_order = [c.lower() for c in other_header.columns_in_order]
        new_inst = cls(
            logical_name=other_header.logical_name,
            primary_key=other_header.primary_key,
            parent=other_header.parent,
            columns_in_order=columns_in_order,
            owner_pid=other_header.owner_pid,
        )
        return new_inst

    def get_column_name(self, input_name: Union[str, Column]) -> str:
        name_str = super().get_column_name(input_name)
        lower_name = name_str.lower()
        self._name_map_db[input_name] = lower_name
        return lower_name

    def has_column(self, column_name) -> bool:
        return self.get_column_name(column_name) in self._columns_positions

    @functools.lru_cache(maxsize=1000)
    def get_column_position(
            self,
            column_name: str,
            allow_create: bool = False,
    ) -> int:
        column_name = self.get_column_name(column_name)
        return super().get_column_position(
            column_name=column_name,
            allow_create=allow_create,
        )

    def row_set_item(
            self,
            column_name: str,
            value,
            row: 'Row',
    ) -> RowIterationHeader:
        column_name_fixed = self.get_column_name(column_name)
        return super().row_set_item(
            column_name=column_name_fixed,
            value=value,
            row=row,
        )

    def rename_column(
            self,
            old_name: str,
            new_name: str,
            ignore_missing: bool = False,
            no_new_header: bool = False,
    ) -> RowIterationHeader:
        """
        Rename a column

        Parameters:
            old_name: str
                The name of the column to find and rename.

            new_name: str
                The new name to give the column.

            ignore_missing: boolean
                Ignore (don't raise error) if we don't have a column with the name in old_name.
                Defaults to False

            no_new_header:
                Skip creating a new row header, modify in place.

                ** BE CAREFUL USING THIS! **

                All new rows created with this header will immediately get the new name,
                in which case you won't want to call this method again.
        """
        old_name = self.get_column_name(old_name)
        new_name = self.get_column_name(new_name)
        return super().rename_column(
            old_name=old_name,
            new_name=new_name,
            ignore_missing=ignore_missing,
            no_new_header=no_new_header,
        )

    def row_remove_column(
            self,
            column_name: str,
            row: 'Row',
            ignore_missing: bool = False,
    ) -> RowIterationHeader:
        column_name = self.get_column_name(column_name)
        return super().row_remove_column(
            column_name=column_name,
            row=row,
            ignore_missing=ignore_missing,
        )



"""
Created on Feb 26, 2015

@author: Derek Wood
"""
# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import functools
import logging
import math
import sys
import traceback
import typing
import warnings
from datetime import datetime

import psutil
from BTrees.OOBTree import OOBTree
from sqlalchemy.sql import Selectable
from sqlalchemy.sql.expression import bindparam

from bi_etl.components.row.row import Row
from bi_etl.components.row.row_iteration_header import RowIterationHeader
from bi_etl.components.row.row_status import RowStatus
from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base
from bi_etl.conversions import ensure_datetime
from bi_etl.exceptions import AfterExisting
from bi_etl.exceptions import BeforeAllExisting
from bi_etl.exceptions import NoResultFound
from bi_etl.memory_size import get_size_gc
from bi_etl.statistics import Statistics

if typing.TYPE_CHECKING:
    from bi_etl.components.etlcomponent import ETLComponent

__all__ = ['Lookup']


class Lookup(object):
    COLLECTION_INDEX = datetime(year=1900, month=1, day=1, hour=0, minute=0, second=0)

    # OOBTree.BTree is now used instead of SortedDict
    # It ran tests in 80% of the time of the SortedDict equivalent
    # http://pythonhosted.org/BTrees/index.html
    VERSION_COLLECTION_TYPE = OOBTree

    ROW_TYPES = typing.Union[Row, typing.Sequence]
    DB_LOOKUP_WARNING = 1000

    def __init__(self,
                 lookup_name: str,
                 lookup_keys: list,
                 parent_component: ETLComponent,
                 config: BI_ETL_Config_Base = None,
                 use_value_cache: bool = True,
                 **kwargs):
        self.lookup_name = lookup_name
        self.lookup_keys = lookup_keys
        self._lookup_keys_set = None
        if config is not None:
            self.config = config
        else:
            self.config = parent_component.task.config
            if self.config is None:
                raise ValueError(f'{self} caller needs to provide config or parent_component that has a config ')
        self._cache = None
        self.parent_component = parent_component
        self.stats = parent_component.get_unique_stats_entry(
            stats_id='{cls} {name}'.format(cls=self.__class__.__name__, name=self.lookup_name)
        )
        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        self._remote_lookup_stmt = None
        self._row_size = None
        self.our_process = psutil.Process()
        self._first_row_process_size = None
        self._row_count_to_get_estimate_row_size = 1000        
        self._done_get_estimate_row_size = False
        self.cache_enabled = True
        self.use_value_cache = use_value_cache
        self._value_cache_str = dict()
        self._value_cache_str_usage = dict()
        self._value_cache_datetime = dict()
        self._value_cache_datetime_usage = dict()
        self.end_date = None
        self._hashable_key_type = tuple
        self._fallback_to_db_count = 0

    def _get_version(
            self,
            versions_collection: Lookup.VERSION_COLLECTION_TYPE[datetime, Row],
            effective_date: datetime
    ):
        # SortedDict version
        # first_effective_index = versions_collection.bisect_right(effective_date) - 1
        # if first_effective_index <= -1:
        #    raise BeforeAllExisting(versions_collection[versions_collection.iloc[0]], effective_date)
        try:
            first_effective_index = versions_collection.maxKey(effective_date)
        except ValueError:
            raise BeforeAllExisting(versions_collection[versions_collection.minKey()], effective_date)
        first_effective_row = versions_collection[first_effective_index]
        # Check that record doesn't end before our date
        if ensure_datetime(first_effective_row[self.end_date]) < effective_date:
            raise AfterExisting(first_effective_row, effective_date)
        return first_effective_row
        
    def __repr__(self):
        return f"{self.__class__.__name__}(lookup_name={self.lookup_name}, lookup_keys={self.lookup_keys})"
    
    @property
    def lookup_keys_set(self):
        if self._lookup_keys_set is None:
            self._lookup_keys_set = set(self.lookup_keys)
        return self._lookup_keys_set

    def _set_log_level(self, log_level):
        self.log.setLevel(log_level)

    def estimated_row_size(self):
        return self._row_size

    def has_done_get_estimate_row_size(self):
        return self._done_get_estimate_row_size

    def get_memory_size(self) -> int:
        if self._row_size is not None:
            return self._row_size * len(self)
        else:
            if len(self) > 0:
                msg = '{lookup_name} Row size was 0 on call to get_memory_size with non zero row count ({cnt:,})'
                msg = msg.format(
                    lookup_name=self.lookup_name,
                    cnt=len(self)
                )
                self.log.warning(msg)
            return 0

    def get_disk_size(self) -> int:
        return 0
    
    def add_size_to_stats(self) -> None:
        self.stats['Final Row Count'] = len(self)
        self.stats['Memory Size'] = self.get_memory_size()
        self.stats['Disk Size'] = self.get_disk_size()

    def get_list_of_lookup_column_values(self, row: ROW_TYPES) -> list:
        if isinstance(row, tuple):
            return list(row)
        elif isinstance(row, list):
            return row
        else:
            lookup_values = [row[k] for k in self.lookup_keys]
            return lookup_values

    @functools.lru_cache(maxsize=10)
    def row_iteration_header_has_lookup_keys(self, row_iteration_header: RowIterationHeader) -> bool:
        for col in self.lookup_keys:
            if col not in row_iteration_header.column_set:
                return False
        return True

    @staticmethod
    def rstrip_key_value(val: object) -> object:
        """
        Since most, if not all, DBs consider two strings that only differ in trailing blanks to
        be equal, we need to rstrip any string values so that the lookup does the same.

        :param val:
        :return:
        """

        if isinstance(val, str):
            return val.rstrip()
        else:
            return val

    def _hashable_key_stripped_non_str(self, row) -> typing.Sequence:
        return self._hashable_key_type(
            tuple(
                [Lookup.rstrip_key_value(val) for val in row]
            )
        )

    def _hashable_key_stripped(self, row) -> typing.Sequence:
        if isinstance(row, str):
            return row
        else:
            return self._hashable_key_stripped_non_str(row)

    def get_hashable_combined_key(self, row: ROW_TYPES) -> typing.Sequence:
        if isinstance(row, self._hashable_key_type):
            if not isinstance(row, str):
                return self._hashable_key_stripped_non_str(row)
            else:
                return row
        else:
            return self._hashable_key_stripped_non_str(self.get_list_of_lookup_column_values(row))

    def clear_cache(self) -> None:
        """
        Removes cache and resets to un-cached state
        """        
        if self._cache is not None:
            del self._cache
        self._cache = None
    
    def __len__(self) -> int:
        if self._cache is not None:
            return len(self._cache)
        else:
            return 0

    def __contains__(self, item):
        return self.has_row(item)

    def init_cache(self) -> None:
        """
        Initializes the cache as empty.
        """
        if self.cache_enabled is None:
            self.cache_enabled = True
        if self.cache_enabled:         
            self._cache = dict()

    def _get_first_row_size(self, row) -> int:
        # get_size_gc is slow but shouldn't be too bad twice (once here and once below in _get_estimate_row_size)
        self._row_size = get_size_gc(row)
        # self.log.debug('First row memory size {:,} bytes (includes overhead)'.format(self._row_size ))
        return self._row_size
        
    def check_estimate_row_size(self, force_now=False):
        if force_now or not self._done_get_estimate_row_size:
            row_cnt = len(self)        
            if force_now or row_cnt >= self._row_count_to_get_estimate_row_size:
                # Tag the estimate as done so we don't keep doing it
                self._done_get_estimate_row_size = True

                # get_size_gc is slow but shouldn't be too bad twice (once here and once above in _get_first_row_size)
                total_cache_size = get_size_gc(self._cache)
                if row_cnt > 0:
                    new_row_size = math.ceil(total_cache_size / row_cnt)
                else:
                    new_row_size = total_cache_size
                self.log.debug(
                    f'{self.lookup_name} Row memory size now estimated at {new_row_size:,} '
                    f'bytes per row using cache of {row_cnt:,} rows'
                )
                self._row_size = new_row_size

    def cache_row(
        self,
        row: Row,
        allow_update: bool = True,
        allow_insert: bool = True,
    ):
        """
        Adds the given row to the cache for this lookup.
        
        Parameters
        ----------
        row: Row
            The row to cache
            
        allow_update: boolean
            Allow this method to update an existing row in the cache.

        allow_insert: boolean
            Allow this method to insert a new row into the cache
            
        Raises
        ------
        ValueError
            If allow_update is False and an already existing row (lookup key) is passed in.
        
        """        
        if self.cache_enabled:
            assert isinstance(row, Row), "cache_row requires Row and not {}".format(type(row))

            if self.use_value_cache:
                self._update_with_value_cache(row)

            lk_tuple = self.get_hashable_combined_key(row)
            if self._cache is None:
                self.init_cache()
            if not allow_update:
                if lk_tuple in self._cache:
                    raise ValueError(
                        f'Key value {lk_tuple} already in cache and allow_update was False.'
                        f' Possible error with the keys defined for this lookup {self.lookup_name} {self.lookup_keys}.'
                    )
            if not allow_insert:
                if lk_tuple not in self._cache:
                    raise ValueError(
                        f'Key value {lk_tuple} not already in cache and allow_insert was False.'
                        f' Error with {self.lookup_name} keys {self.lookup_keys}.'
                    )
            self._cache[lk_tuple] = row
            
            # Capture memory usage snapshots
            if self._row_size is None:   
                self._get_first_row_size(row)
            else:
                self.check_estimate_row_size()

    def cache_set(
            self,
            lk_tuple: tuple,
            version_collection: OOBTree[datetime, Row],
            allow_update: bool = True
            ):
        """
        Adds the given set of rows to the cache for this lookup.

        Parameters
        ----------
        lk_tuple:
            The key tuple to store the rows under

        version_collection:
            The set of rows to cache

        allow_update: boolean
            Allow this method to update an existing row in the cache.

        Raises
        ------
        ValueError
            If allow_update is False and an already existing row (lookup key) is passed in.

        """
        if self.cache_enabled:
            assert isinstance(version_collection, Lookup.VERSION_COLLECTION_TYPE), \
                f"cache_row requires {Lookup.VERSION_COLLECTION_TYPE} and not {type(version_collection)}"

            if self.use_value_cache:
                for row in version_collection.values():
                    self._update_with_value_cache(row)

            if self._cache is None:
                self.init_cache()
            if not allow_update:
                if lk_tuple in self._cache:
                    raise ValueError('Key value {} already in cache and allow_update was False.'
                                     ' Possible error with the keys defined for this lookup {} {}.'
                                     .format(lk_tuple, self.lookup_name, self.lookup_keys)
                                     )
            self._cache[lk_tuple] = version_collection

            # Capture memory usage snapshots
            if self._row_size is None:
                self._get_first_row_size(version_collection)
            else:
                self.check_estimate_row_size()

    def commit(self):
        """
        Placeholder for other implementations that might need it
        """
        pass
        
    def uncache_row(self, row: ROW_TYPES):
        if self._cache is not None:
            try:
                lk_tuple = self.get_hashable_combined_key(row)
            except KeyError as e:
                # This lookup uses columns not in the row provided.
                raise ValueError(f"uncache_row called on {self} with insufficient values {row}. {e}")
            try:
                del self._cache[lk_tuple]
            except KeyError:
                pass

    def uncache_set(self, row: ROW_TYPES):
        self.uncache_row(row)

    def __iter__(self) -> typing.Iterable[Row]:
        """
        Iterates over rows in the lookup cache.  Returns clones of rows in case they are modified.

        The rows will come out in any order.
        """
        if self._cache is not None:
            for row in list(self._cache.values()):
                # Do not return the actual instance of the row since the caller might modify it to no longer be
                # compatible with the parent table
                yield row.clone()

    def _iter_raw(self) -> typing.Iterable[Row]:
        """
        Iterates over rows in the lookup cache. Returns actual cached rows. Be careful! Direct modifications to
        the cached rows could break the later use of the cache.

        The rows will come out in any order.
        :return:
        """
        if self._cache is not None:
            for row in self._cache.values():
                yield row

    def find_where(
            self,
            key_names: typing.Sequence,
            key_values_dict: typing.Mapping,
            limit: int = None
            ):
        """
        Scan all cached rows (expensive) to find list of rows that match criteria.
        """
        results = list()
        for row in self:
            matches = True
            for key in key_names:
                if row[key] != key_values_dict[key]:
                    matches = False
                    break
            if matches:
                results.append(row)
                if limit is not None and len(results) >= limit:
                    break
        return results
            
    def uncache_where(
            self,
            key_names: typing.Sequence,
            key_values_dict: typing.Mapping):
        """
        Scan all cached rows (expensive) to find rows to remove.
        """
        deletes = self.find_where(key_names, key_values_dict)
        for row in deletes:
            self.uncache_row(row)

    def get_versions_collection(
            self,
            row: ROW_TYPES
            ) -> typing.MutableMapping[datetime, Row]:
        """
        This method exists for compatibility with range caches

        Parameters
        ----------
        row
            The row with keys to search row

        Returns
        -------
        A MutableMapping of rows
        """
        if not self.cache_enabled:
            raise ValueError("Lookup {} cache not enabled".format(self.lookup_name))
        if self._cache is None:
            self.init_cache()

        lk_tuple = self.get_hashable_combined_key(row)
        try:
            return Lookup.VERSION_COLLECTION_TYPE({Lookup.COLLECTION_INDEX: self._cache[lk_tuple]})
        except KeyError as e:
            raise NoResultFound(e)

    def find_in_cache(self, row: ROW_TYPES, **kwargs) -> Row:
        """
        Find a matching row in the lookup based on the lookup index (keys)
        """
        assert len(kwargs) == 0, "lookup.find_in_cache got unexpected args {}".format(kwargs)
        versions_collection = self.get_versions_collection(row)
        return versions_collection[Lookup.COLLECTION_INDEX]

    def find_matches_in_cache(self, row: ROW_TYPES, **kwargs) -> typing.Sequence[Row]:
        return list(self.find_in_cache(row, **kwargs))

    def has_row(self, row: ROW_TYPES) -> bool:
        """
        Does the row exist in the cache (for any date if it's a date range cache)

        Parameters
        ----------
        row

        Returns
        -------

        """
        try:
            self.get_versions_collection(row)
        except (NoResultFound, BeforeAllExisting, AfterExisting):
            return False
        return True
            
    def _add_remote_stmt_where_clause(self, stmt: Selectable) -> Selectable:
        """
        Only works if parent_component is based on bi_etl.components.readonlytable
        """
        col_num = 1
        for key_col in self.lookup_keys:
            # This statement only works if parent_component is based on bi_etl.components.readonlytable
            # noinspection PyUnresolvedReferences
            stmt = stmt.where(self.parent_component.get_column(key_col) == bindparam('k{}'.format(col_num)))
            col_num += 1
        return stmt

    def _get_remote_stmt_where_values(
            self,
            row: ROW_TYPES,
            effective_date: datetime = None,
            ) -> dict:
        values_dict = dict()
        col_num = 1
        values_list = self.get_list_of_lookup_column_values(row)

        for key_val in values_list:
            values_dict['k{}'.format(col_num)] = key_val
            col_num += 1
        return values_dict

    def find_in_remote_table(
            self,
            row: ROW_TYPES,
            **kwargs
            ) -> Row:
        """
        Find a matching row in the lookup based on the lookup index (keys)
        
        Only works if parent_component is based on bi_etl.components.readonlytable

        Parameters
        ----------
        row
            The row with keys to search row

        Returns
        -------
        A row
        """
        assert hasattr(self.parent_component, 'execute'), \
            f'ReadOnlyTable or class with execute method needed ' \
            f'for DB lookup {type(self.parent_component)} will not work'

        self.stats.timer.start()        
        if self._remote_lookup_stmt is None:
            # noinspection PyUnresolvedReferences
            stmt = self.parent_component.select()
            stmt = self._add_remote_stmt_where_clause(stmt)
            self._remote_lookup_stmt = stmt.compile()

        values_dict = self._get_remote_stmt_where_values(row)

        # noinspection PyUnresolvedReferences
        rows = list(self.parent_component.execute(self._remote_lookup_stmt, values_dict))
        self.stats.timer.stop()
        if len(rows) == 0:
            raise NoResultFound()
        elif len(rows) == 1:
            # noinspection PyUnresolvedReferences
            return self.parent_component.Row(data=rows[0])
        else:
            msg = "{lookup_name} find_in_remote_table {row} matched multiple records {rows}"
            msg = msg.format(lookup_name=self.lookup_name,
                             row=row,
                             rows=rows
                             )
            raise RuntimeError(msg)

    def find_versions_list_in_remote_table(self, row: ROW_TYPES) -> list:
        return [self.find_in_remote_table(row)]

    def find_versions_list(
            self,
            row: ROW_TYPES,
            fallback_to_db: bool = True,
            maintain_cache: bool = True,
            stats: Statistics = None,
            ) -> list:
        """

        Parameters
        ----------
        row:
            row or tuple to find
        fallback_to_db:
            Use db to search if not found in cached copy
        maintain_cache:
            Add DB lookup rows to the cached copy?
        stats:
            Statistics to maintain

        Returns
        -------
        A MutableMapping of rows
        """
        if self.cache_enabled:
            try:
                rows = list(self.get_versions_collection(row).values())
                if stats is not None:
                    stats['Found in cache'] += 1
                    stats.timer.stop()
                return rows
            except NoResultFound as e:
                if stats is not None:
                    stats['Not in cache'] += 1
                if not fallback_to_db:
                    #  Don't pass onto SQL if the lookup cache has initialized but the value isn't there
                    if stats is not None:
                        stats.timer.stop()
                    raise e
                    # Otherwise we'll continue to the DB query below
            except (KeyError, ValueError):
                if stats is not None:
                    stats['Cache not setup'] += 1
                # Lookup not cached. Allow database lookup to proceed, but give warning since we thought it was cached
                warnings.warn("WARNING: {lkp} caching is enabled, but find_in_cache returned error {e}".format(
                    lkp=self,
                    e=traceback.format_exc()
                )
                )
            except AfterExisting as e:
                stats['After all in cache'] += 1
                if stats is not None:
                    stats['Not in cache'] += 1
                if not fallback_to_db:
                    #  Don't pass onto SQL if the lookup cache has initialized but the value isn't there
                    stats.timer.stop()
                    # return the last existing row since that what we'll update
                    return e.prior_row
                    # Otherwise we'll continue to the DB query below
                    # BeforeExisting exceptions are returned to the caller

        # Do a lookup on the database
        try:
            if stats is not None:
                stats['DB lookup performed'] += 1
            rows = self.find_versions_list_in_remote_table(row)
            for row in rows:
                row.status = RowStatus.existing
                if maintain_cache and self.cache_enabled:
                    self.cache_row(row, allow_update=False)
                if stats is not None:
                    stats['DB lookup found row'] += 1
        finally:
            if stats is not None:
                stats.timer.stop()
        return rows

    def find(self,
             row: ROW_TYPES,
             fallback_to_db: bool = True,
             maintain_cache: bool = True,
             stats: Statistics = None,
             **kwargs
             ) -> Row:
        if self.cache_enabled:
            try:
                row = self.find_in_cache(row, **kwargs)
                if stats is not None:
                    stats['Found in cache'] += 1
                    stats.timer.stop()
                return row
            except NoResultFound as e:
                if stats is not None:
                    stats['Not in cache'] += 1
                if fallback_to_db:
                    self._fallback_to_db_count += 1
                    if (self._fallback_to_db_count % self.DB_LOOKUP_WARNING) == 0:
                        self.log.warning(
                            f"{self} has done {self._fallback_to_db_count} lookups on the DB due to fallback_to_db"
                        )
                else:
                    #  Don't pass onto SQL if the lookup cache has initialized but the value isn't there
                    if stats is not None:
                        stats.timer.stop()
                    raise e
                    # Otherwise we'll continue to the DB query below
            except (KeyError, ValueError):
                if stats is not None:
                    stats['Cache not setup'] += 1
                # Lookup not cached. Allow database lookup to proceed, but give warning since we thought it was cached
                warnings.warn("WARNING: {lkp} caching is enabled, but find_in_cache returned error {e}".format(
                    lkp=self,
                    e=traceback.format_exc()
                )
                )
            except AfterExisting as e:
                stats['After all in cache'] += 1
                if stats is not None:
                    stats['Not in cache'] += 1

                if fallback_to_db:
                    if (stats['Not in cache'] % self.DB_LOOKUP_WARNING) == 0:
                        self.log.warning(
                            f"{self} has done {stats['Not in cache']} lookups on the DB due to fallback_to_db")
                else:
                    #  Don't pass onto SQL if the lookup cache has initialized but the value isn't there
                    stats.timer.stop()
                    # return the last existing row since that what we'll update
                    return e.prior_row
                    # Otherwise we'll continue to the DB query below
                    # BeforeExisting exceptions are returned to the caller

        # Do a lookup on the database
        try:
            if stats is not None:
                stats['DB lookup performed'] += 1
            row = self.find_in_remote_table(row, **kwargs)
            row.status = RowStatus.existing
            if maintain_cache and self.cache_enabled:
                self.cache_row(row, allow_update=False)
            if stats is not None:
                stats['DB lookup found row'] += 1
        finally:
            if stats is not None:
                stats.timer.stop()
        return row

    def _update_with_value_cache(self, row):
        """
        Update a row with str and datetime de-duplication references

        :param row:
        :return:
        """
        for column_number, column_value in enumerate(row.values_in_order()):
            if isinstance(column_value, str):
                if column_value in self._value_cache_str:
                    # Replace with value cache reference
                    row.set_by_zposition_unsafe(column_number, self._value_cache_str[column_value])
                    self._value_cache_str_usage[column_value] += 1
                else:
                    # Add to value cache
                    self._value_cache_str[column_value] = column_value
                    self._value_cache_str_usage[column_value] = 0
            elif isinstance(column_value, datetime):
                if column_value in self._value_cache_datetime:
                    # Replace with value cache reference
                    row.set_by_zposition_unsafe(column_number, self._value_cache_datetime[column_value])
                    self._value_cache_datetime_usage[column_value] += 1
                else:
                    # Don't bother adding microsecond level datetime values to the dict
                    # It will perform better if smaller
                    if column_value.microsecond in {0, 999990, 999999}:
                        # Add to value cache
                        self._value_cache_datetime[column_value] = column_value
                        self._value_cache_datetime_usage[column_value] = 0

    @staticmethod
    def _str_report_on_value_cache_usage(lookup_name: str, type_name: str, value_cache: dict) -> str:
        key_count = len(value_cache)
        if key_count > 0:
            max_usage = max(value_cache.values())
            total_usage = sum(value_cache.values())
            total_bytes_saved = sum([sys.getsizeof(key) * value for key, value in value_cache.items()])
            return f'{lookup_name} {type_name} value cache stored {key_count:,} values. ' \
                   f'Max key usage = {max_usage:,}. ' \
                   f'Avg key usage = {100.0 * total_usage / key_count :.2f}%. total_bytes_saved = {total_bytes_saved:,}'
        else:
            return f'{lookup_name} {type_name} value cache stored no values in cache'

    def report_on_value_cache_effectiveness(self, lookup_name: str = None):
        if lookup_name is None:
            lookup_name = self.lookup_name
        self.log.info(self._str_report_on_value_cache_usage(lookup_name, 'str', self._value_cache_str_usage))
        self.log.info(self._str_report_on_value_cache_usage(lookup_name, 'datetime', self._value_cache_datetime_usage))



import email
import re
import smtplib
from email.mime.text import MIMEText

import bi_etl.config.notifiers_config as notifiers_config
from bi_etl.notifiers.notifier_base import NotifierBase, NotifierException


class Email(NotifierBase):
    def __init__(self, config_section: notifiers_config.SMTP_Notifier):
        super().__init__()
        self.config_section = config_section

    def send(self, subject, message, sensitive_message=None, attachment=None, throw_exception=False):
        distro_list = self.config_section.distro_list
        if not distro_list:
            self.log.warning(f'{self.config_section} distro_list option not found. No mail sent.')

        to_addresses = list()
        if isinstance(distro_list, list):
            to_addresses.extend(distro_list)
        elif isinstance(distro_list, str):
            for to_address in re.split(r'[,;\n]', distro_list):
                to_address = to_address.strip()
                self.log.info('Adding {} to send list'.format(to_address))
                to_addresses.append(to_address)
        else:
            raise ValueError(f"distro_list not list or string but {type(distro_list)} with value {distro_list}")

        server = None
        try:
            if isinstance(message, email.message.Message):
                if subject is not None:
                    message['subject'] = subject
                if 'To' not in message:
                    message['To'] = ','.join(to_addresses)
                if 'From' not in message:
                    if 'Sender' not in message:
                        message['Sender'] = self.config_section.email_from
            else:
                if message is None:
                    message = ''
                message = MIMEText(message)
                if subject is not None:
                    subject_escaped = subject
                    reserved_list = ['\n', '\r']
                    for reserved in reserved_list:
                        subject_escaped = subject_escaped.replace(reserved, ' ')

                    message['subject'] = subject_escaped
                message['Sender'] = self.config_section.email_from
                message['To'] = ','.join(to_addresses)

            gateway = self.config_section.gateway_host
            gateway_port = self.config_section.gateway_port
            gateway_userid = self.config_section.user_id
            gateway_password = self.config_section.get_password()

            use_ssl = self.config_section.use_ssl
            if use_ssl:
                server = smtplib.SMTP_SSL(gateway, port=gateway_port)
            else:
                server = smtplib.SMTP(gateway, port=gateway_port)
            server.set_debuglevel(self.config_section.debug)
            if gateway_userid is not None:
                server.login(gateway_userid, gateway_password)

            results_of_send = server.send_message(message)
            self.log.debug("results_of_send = {}".format(results_of_send))

            for recipient in results_of_send:
                self.log.warn(f"Problem sending to: {recipient}")
        except smtplib.SMTPRecipientsRefused as e:
            self.log.critical(f"All recipients were refused.\n{e.recipients}")
            if throw_exception:
                raise NotifierException(e)
        except smtplib.SMTPHeloError as e:
            self.log.critical(f"The server didn't reply properly to the HELO greeting.\n{e}")
            if throw_exception:
                raise NotifierException(e)
        except smtplib.SMTPSenderRefused as e:
            self.log.critical(f"The server didn't accept the from_addr {message.get('Sender', None)}.\n{e}")
            if throw_exception:
                raise NotifierException(e)
        except smtplib.SMTPDataError as e:
            self.log.critical(
                f"The server replied with an unexpected error code (other than a refusal of a recipient).\n{e}"
            )
            if throw_exception:
                raise NotifierException(e)
        finally:
            try:
                if server is not None:
                    reply = server.quit()
                    self.log.debug(f"server quit reply = {reply}")
                    self.log.info('Mail sent')
            except Exception as e:
                self.log.exception(e)



import bi_etl.config.notifiers_config as notifiers_config
from bi_etl.notifiers.notifier_base import NotifierBase


class Jira(NotifierBase):
    def __init__(self, config_section: notifiers_config.JiraNotifier):
        super().__init__()
        self.config_section = config_section

        # On-instance import since jira is an optional requirement
        # noinspection PyUnresolvedReferences
        from jira.client import JIRA
        # noinspection PyUnresolvedReferences
        from jira.exceptions import JIRAError

        self.config_section = config_section
        options = dict()
        options['server'] = self.config_section.server
        user_id = self.config_section.user_id
        self.project = self.config_section.project
        password = self.config_section.get_password()
        self.log.debug(f"user id={user_id}")
        self.log.debug(f"server={options['server']}")
        self.log.debug(f'project={self.project}')

        try:
            self.jira_conn = JIRA(options, basic_auth=(user_id, password))
        except JIRAError as e:
            if 'CAPTCHA_CHALLENGE' in e.text:
                raise RuntimeError(f'Jira Login requests passing CAPTCHA CHALLENGE.  {e.text}')
            else:
                self.log.error(f'Error connecting to JIRA')
                self.log.exception(e)
                raise
        priority_name = self.config_section.priority
        if priority_name is not None:
            for priority_object in self.jira_conn.priorities():
                if priority_object.name == priority_name:
                    self.priority_id = priority_object.id
            self.log.debug(f'priority_name = {priority_name} priority_id={self.priority_id}')
        else:
            self.priority_id = None
            self.log.debug('priority not specified in config')

        self.subject_prefix = self.config_section.subject_prefix
        self.comment_on_each_instance = self.config_section.comment_on_each_instance
        self.component = self.config_section.component
        self.issue_type = self.config_section.issue_type
        exclude_statuses = self.config_section.exclude_statuses
        exclude_statuses_filter_list = []
        for status in exclude_statuses:
            exclude_statuses_filter_list.append(f'"{status}"')
        self.exclude_statuses_filter = ','.join(exclude_statuses_filter_list)

    def add_attachment(self, issue, attachment):
        """Attach an attachment to an issue and returns a Resource for it.

        The client will *not* attempt to open or validate the attachment; it expects a file-like object to be ready
        for its use. The user is still responsible for tidying up (e.g., closing the file, killing the socket, etc.)

        :param issue: the issue to attach the attachment to
        :param attachment:
            file-like object to attach to the issue, also works if it is a string with the filename,
            or a tuple with a file-like object and a filename.

            If the (file, filename) tuple is not used the file object's ``name`` attribute
            is used. If you acquired the file-like object by any other method than ``open()``, make sure
            that a name is specified in one way or the other.
        :rtype: an Attachment Resource
        """
        if isinstance(attachment, tuple):
            attachment, filename = attachment
        else:
            filename = None

        return self.jira_conn.add_attachment(issue=issue, attachment=attachment, filename=filename)

    def search(self, subject):
        # Find already opened case, if there is one
        found_issues = list()
        # Remove any special characters that break JQL parsing
        # https://support.atlassian.com/jira-software-cloud/docs/search-syntax-for-text-fields/
        subject_escaped = subject
        reserved_list = [
            '\\', '+', '-', '[', ']', '(', ')', '{', '}',
            'AND', 'OR', 'NOT',
            '"', "'", '|', '&&', '!', '*', ':',
            '?', '~', '^', '%',
            '\t', '\n', '\r',
        ]
        for reserved in reserved_list:
            subject_escaped = subject_escaped.replace(reserved, ' ')

        issues = self.jira_conn.search_issues(
            f'project="{self.project}" '
            f'AND summary~"{subject_escaped}" '
            f'AND status not in ({self.exclude_statuses_filter})'
        )
        for iss in issues:
            # Double check that name matches since JIRA does a wildcard search and word stemming
            if iss.fields.summary.strip() == subject:
                # self.log.debug('Potential match:')
                # self.log.debug(p.fields.status)
                # self.log.debug(p.fields.summary)
                # self.log.debug(p.fields.description)
                found_issues.append(iss)
                case_number = iss.key
        return found_issues

    def send(self, subject, message, sensitive_message=None, attachment=None, throw_exception=False):
        """
        Log a Jira issue

        To use special formatting codes plesae see
        https://jira.atlassian.com/secure/WikiRendererHelpAction.jspa?section=all

        :param subject:
        :param message:
        :param sensitive_message:
        :param attachment:
        :param throw_exception:
        :return:
        """
        if subject is None:
            raise ValueError(f"Jira notifier requires a valid subject. Message was {message}")
        else:
            subject = self.subject_prefix + subject.strip()
        self.log.debug(f'subject={subject}')
        self.log.debug(f'message={message}')

        if message is None:
            message = "_No Description Provided_"

        message_parts = [
            message,
        ]
        if sensitive_message is not None:
            message_parts.append(sensitive_message)

        existing_issues = self.search(subject)
        if len(existing_issues) > 1:
            self.log.warning(f"Found multiple open issues with subject {subject}. Finding newest...")
            newest_case_number = 0
            newest_iss = None
            for iss in existing_issues:
                case_number = iss.key
                self.log.info(f"One of multiple existing open cases is {case_number}.")
                # Fixed the issue in the file by getting the int value for case_number
                proj_code, case_num = case_number.split('-')
                case_num_int = int(case_num)
                if case_num_int > newest_case_number:
                    newest_case_number = case_num_int
                    newest_iss = iss
            existing_issues = [newest_iss]
            # Allow the section below to comment on the newest issue

        if len(existing_issues) == 1:
            iss = existing_issues[0]
            case_number = iss.key
            self.log.info(f"Found existing open case {case_number}.")
            if self.comment_on_each_instance:
                if attachment is not None:
                    attachment_object = self.add_attachment(iss, attachment)
                    self.log.debug(f"Created attachment {attachment_object}")
                message_parts.insert(0, "New occurrence with message(s):")
                if message or sensitive_message:
                    comment = '\n'.join(message_parts)
                    self.jira_conn.add_comment(iss, comment)
                    self.log.info(f"Added comment to case {case_number}.")
        else:
            description = '\n'.join(message_parts)

            issue_dict = {
                'project': {'key': self.project},
                'summary': subject,
                'description': description,
            }
            if self.issue_type is not None:
                issue_dict['issuetype'] = {'name': self.issue_type}
            if self.priority_id:
                issue_dict['priority'] = {'id': self.priority_id}
            if self.component:
                issue_dict['components'] = [{'name': self.component}, ]

            self.log.debug('issue_dict={}'.format(issue_dict))

            new_issue = self.jira_conn.create_issue(fields=issue_dict)
            case_number = new_issue.key
            self.log.info("Created new case {}".format(case_number))

            if attachment is not None:
                attachment_object = self.add_attachment(new_issue, attachment)
                self.log.debug(f"Created attachment {attachment_object}")



"""
Created on Sep 17, 2014

"""
import csv
import logging
import os
import typing

from sqlalchemy import Column

from bi_etl.components.etlcomponent import ETLComponent
from bi_etl.components.row.row import Row
from bi_etl.scheduler.task import ETLTask
from bi_etl.statistics import Statistics

__all__ = ['W3CReader']

# Only quote none is really needed here, QUOTE_MINIMAL is the default.
# The other quoting levels are only relevant to the Writer
QUOTE_NONE = csv.QUOTE_NONE
QUOTE_MINIMAL = csv.QUOTE_MINIMAL


class W3CReader(ETLComponent):
    """
    W3CReader reads W3c based log files

    Args:
        task: The  instance to register in (if not None)

        filedata:
            The file to parse as delimited. If str then it's assumed to be a filename.
            Otherwise it's assumed to be a file object.

        encoding:
            The encoding to use when opening the file,
            if it was a filename and not already opened.
            Default is None which becomes the Python default encoding

        errors:
            The error handling to use when opening the file
            (if it was a filename and not already opened)
            Default is 'strict'
            See above for valid errors values.

        logical_name:
            The logical name of this source. Used for log messages.

    Attributes:
        column_names: list
            The names to use for columns
    """

    def __init__(self,
                 task: ETLTask,
                 filedata: typing.Union[typing.TextIO, str],
                 encoding: str = None,
                 errors: str = 'strict',
                 logical_name: str = None,
                 **kwargs
                 ):

        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")

        self.__close_file = False

        # We have to check / open the file here to get the name for the logical name
        if isinstance(filedata, str):
            self.log.info("Opening file {}".format(filedata))
            self.file = open(filedata,
                             mode='rt',
                             newline='',
                             encoding=encoding,
                             errors=errors
                             )
            self.__close_file = True
        else:
            self.log.info("Treating input as file object {}".format(filedata))
            self.file = filedata

        if logical_name is None:
            try:
                logical_name = os.path.basename(self.file.name)
            except AttributeError:
                logical_name = str(self.file)

        self.read_all = True
        self._file_read = False

        self._line_num = 0
        self._pending_lines = list()

        # Don't pass kwargs up. They should be set here at the end
        super().__init__(
            task=task,
            logical_name=logical_name,
        )

        # Should be the last call of every init
        self.set_kwattrs(**kwargs)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(task={self.task},logical_name={self.logical_name}," 
            f"filedata={self.file},primary_key={self.primary_key},column_names={self.column_names})"
        )

    @property
    def line_num(self):
        """
        The current line number in the source file.
        line_num differs from rows_read in that rows_read deals with rows that would be returned to the caller
        """
        return self._line_num

    def _read_line(self) -> typing.Union[str, None]:
        if len(self._pending_lines) == 0:
            if self.read_all:
                self._pending_lines = self.file.readlines()
                if self._pending_lines:
                    line = self._pending_lines.pop(0)
                    self._line_num += 1
                    return line.strip()
                else:
                    return None
            else:
                line = self.file.readline()
                if line:
                    self._line_num += 1
                return line.strip()
        else:
            return self._pending_lines.pop(0).strip()

    def _read_line_no_comments(self) -> typing.Union[str, None]:
        line = self._read_line()
        while line is not None and line[0] == '#':
            line = self._read_line()
        return line

    def _obtain_column_names(self):
        """
        Get the column names from the file. ETLComponent only call this if self._column_names is None:
        """
        header_found = False
        while not header_found:
            line = self._read_line()
            if line is None:
                raise ValueError("EOF found before column names")
            if line.startswith('#Fields:'):
                self.column_names = line.split(' ')[1:]
                header_found = True
            elif not line.startswith('#'):
                raise ValueError("Data lines started before column names")
        if self.trace_data:
            self.log.debug("Column names read: {}".format(self.column_names))

    def where(
            self,
            criteria_list: typing.Optional[list] = None,
            criteria_dict: typing.Optional[dict] = None,
            order_by: typing.Optional[list] = None,
            column_list: typing.List[typing.Union[Column, str]] = None,
            exclude_cols: typing.List[typing.Union[Column, str]] = None,
            use_cache_as_source: typing.Optional[bool] = None,
            progress_frequency: typing.Optional[int] = None,
            stats_id: typing.Optional[str] = None,
            parent_stats: typing.Optional[Statistics] = None,
    ) -> typing.Iterable[Row]:
        """

        Parameters
        ----------
        criteria_list:
            *Not Supported for this component.*
        criteria_dict:
            Dict keys should be column names, values are checked for equality with the column on each row.
        order_by:
            *Not Supported for this component.*
        column_list:
            List of columns names
        exclude_cols
        use_cache_as_source
        progress_frequency
        stats_id
        parent_stats

        Returns
        -------
        rows

        """
        return super().where(
            criteria_list=criteria_list,
            criteria_dict=criteria_dict,
            order_by=order_by,
            column_list=column_list,
            exclude_cols=exclude_cols,
            use_cache_as_source=use_cache_as_source,
            progress_frequency=progress_frequency,
            stats_id=stats_id,
            parent_stats=parent_stats,
        )

    def _raw_rows(self):
        len_column_names = len(self.column_names)
        try:
            this_iteration_header = self.full_iteration_header
            # noinspection PyTypeChecker
            done = False
            while not done:
                line = self._read_line_no_comments()
                if line is None:
                    done = True
                else:
                    if line != '':
                        # Convert empty strings to None to be consistent with DB reads
                        row = line.split(' ')
                        d = self.Row(data=row[:len_column_names], iteration_header=this_iteration_header)

                        len_row = len(row)
                        if len_column_names < len_row:
                            raise ValueError(
                                f'Row {self.line_num} has extra columns Row has {len_row} > Header has {len_column_names}: "{line}"')
                        elif len_column_names > len_row:
                            if self.warnings_issued < self.warnings_limit:
                                self.warnings_issued += 1
                                self.log.warning(
                                    f'Row {self.line_num} is missing columns. Row has {len_row} < Header has {len_column_names}: "{line}"')
                            for key in self.column_names[len_row:]:
                                d[key] = None
                        yield d
        finally:
            pass

    def close(self, error: bool = False):
        """
        Close the file
        """
        if not self.is_closed:
            if self.__close_file:
                self.file.close()
            super().close(error=error)



"""
Created on March 28, 2022

@author: 
"""
import logging
import os
import time
from datetime import datetime, timezone
from enum import Enum
from typing import *

import botocore
from config_wrangler.config_templates.aws.s3_bucket import S3_Bucket

from bi_etl.database import DatabaseMetadata


class File_Entry(object):
    def __init__(
            self,
            content_path: str,
            full_local_path: str,
            source_path: str = None,
            s3_last_modified: datetime = None,
            s3_file_size: int = None,
            bucket_object=None,
            local_last_modified: datetime = None,
            local_file_size: int = None,
    ):
        self.content_path = content_path
        self.full_local_path = full_local_path
        self.source_path = source_path
        self.s3_last_modified = s3_last_modified
        self.s3_file_size = s3_file_size
        self.bucket_object = bucket_object
        self.local_last_modified = local_last_modified
        self.local_file_size = local_file_size

    def __str__(self):
        return self.content_path

    def __repr__(self):
        return f'{self.content_path} ({self.s3_file_size:,} {self.s3_last_modified})'

    def full_path(self, base_path):
        return os.path.join(base_path, self.content_path)

    def basename(self):
        return os.path.basename(self.content_path)


class FileFormat(Enum):
    parquet = 'P'
    csv = 'C'


class Boto3_S3(S3_Bucket):

    def download_file_from_s3(
            self,
            file_key: str,
            local_file_path: str,
            date_placeholder: str = None,
            date_pattern: str = '%Y-%m-%d',
    ) -> str:
        log = logging.getLogger(__name__)

        if date_placeholder:
            # Get the timestamp of the file
            s3_object = self.client.get_object(Bucket=self.bucket_name, Key=file_key)

            file_date = s3_object['LastModified']
            log.info(f"{file_key} timestamp: {file_date}")
            file_date_str = file_date.strftime(date_pattern)

            local_file_path = local_file_path.format(**{date_placeholder: file_date_str})

        done = False
        tries = 0
        while not done:
            try:
                # Download the file
                self.resource.Bucket(self.bucket_name).download_file(
                    str(file_key),
                    str(local_file_path)
                )

                return local_file_path

            except botocore.exceptions.ClientError as e:
                if e.response['Error']['Code'] == "404":
                    log.error(f"The object does not exist. {self}/{file_key}")
                    raise e
                else:
                    log.error(f"Error downloading {self}/{file_key} was {e}")
                    raise e
            except OSError as e:
                if '.' in local_file_path and tries < 3:
                    local_file_path = local_file_path.replace('.', '_.')
                    log.warning(f"Got error {e} will try again with name '{local_file_path}'")
                    tries += 1
                else:
                    raise

    def scan_files_from_s3(
            self,
            key_prefixes: List[str],
            local_folder: Optional[str] = None,
    ) -> List[File_Entry]:
        log = logging.getLogger(__name__)

        file_list = list()
        local_timezone = datetime.now(timezone.utc).astimezone().tzinfo

        try:
            bucket = self.resource.Bucket(self.bucket_name)
            for key_prefix in key_prefixes:
                key_prefix = key_prefix.strip()
                if key_prefix != '':
                    log.info(f'Scanning files in key {key_prefix}')
                    for bucket_object in bucket.objects.filter(Prefix=key_prefix):
                        file_name = bucket_object.key
                        if len(file_name) > 0 and file_name[-1] not in {'/', '\\'}:
                            local_file_path = None
                            local_last_modified = None
                            local_file_size = None
                            if local_folder:
                                local_file_path = os.path.join(local_folder, file_name)
                                if os.path.exists(local_file_path):
                                    local_last_modified = datetime.fromtimestamp(
                                        os.path.getmtime(local_file_path),
                                        tz=local_timezone
                                    )
                                    local_file_size = os.path.getsize(local_file_path)

                            file_entry = File_Entry(
                                content_path=bucket_object.key,
                                full_local_path=local_file_path,
                                source_path=f'{self.bucket_name}/{bucket_object.key}',
                                s3_last_modified=bucket_object.last_modified,
                                s3_file_size=bucket_object.size,
                                bucket_object=bucket_object,
                                local_last_modified=local_last_modified,
                                local_file_size=local_file_size,
                            )

                            file_list.append(file_entry)
            return file_list

        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                log.error("The object does not exist.")
                raise e
            else:
                raise e

    @staticmethod
    def download_file_list_from_s3(
            file_list: List[File_Entry]
    ):
        log = logging.getLogger(__name__)

        for file_entry in file_list:
            log.info(f'Downloading {file_entry.content_path}')
            full_dir_path = os.path.dirname(file_entry.full_local_path)
            os.makedirs(full_dir_path, exist_ok=True)
            body = file_entry.bucket_object.get()['Body']

            with open(file_entry.full_local_path, 'wb') as local_file:
                while local_file.write(body.read(amt=512)):
                    pass

    @staticmethod
    def filter_list_changed_vs_local(file_list: List[File_Entry]) -> List[File_Entry]:
        result_list = list()
        for file_entry in file_list:
            if file_entry.local_last_modified is None:
                result_list.append(file_entry)
            elif file_entry.local_last_modified < file_entry.s3_last_modified:
                result_list.append(file_entry)
            elif file_entry.local_file_size is None:
                result_list.append(file_entry)
            elif file_entry.local_file_size != file_entry.s3_file_size:
                result_list.append(file_entry)
        return result_list

    def download_files_from_s3(
            self,
            key_prefixes: List[str],
            local_folder: str,
            changed_only=True,
            return_full_list=True,
    ) -> List[File_Entry]:

        file_list = self.scan_files_from_s3(
            key_prefixes=key_prefixes,
            local_folder=local_folder,
        )

        if changed_only:
            download_file_list = Boto3_S3.filter_list_changed_vs_local(file_list)
            if not return_full_list:
                file_list = download_file_list
        else:
            download_file_list = file_list

        Boto3_S3.download_file_list_from_s3(download_file_list)

        return file_list

    def upload_file_to_s3(
            self,
            file_key: str,
            local_file_path: str,
    ):
        log = logging.getLogger(__name__)

        done = False
        tries = 0
        while not done:
            try:
                # Upload the file
                self.client.upload_file(
                    Filename=str(local_file_path),
                    Bucket=self.bucket_name,
                    Key=str(file_key),
                )
                done = True

            except botocore.exceptions.ClientError as e:
                if e.response['Error']['Code'] == "404":
                    log.error("The object does not exist.")
                    raise e
                else:
                    raise e
            except OSError as e:
                if tries < 3:
                    log.warning(f"Got error {e} will try again with name '{local_file_path}'")
                    tries += 1
                else:
                    raise

    def unload_data(
            self,
            database: DatabaseMetadata,
            file_format: FileFormat = None,
            query: str = None,
            out_folder: str = None,
            filename: str = None,
            delimiter: str = None,
        ):

        log = logging.getLogger(__name__)
        counter = 0

        filename = filename.replace('\\', '/')
        fldr = out_folder.replace('\\', '/')

        sql_source = f"""
            UNLOAD ('{query}') to 's3://{self.bucket_name}/{fldr}/' 
            credentials 'aws_access_key_id={self.user_id};aws_secret_access_key={self.get_password()}'
            parallel off ALLOWOVERWRITE 
            REGION '{self.region_name}'
        """

        if file_format is FileFormat.csv:
            sql_source += f" DELIMITER '{delimiter}' CSV HEADER;"
            filenm = '000'
        elif file_format is FileFormat.parquet:
            sql_source += f" PARQUET;"
            filenm = '000.parquet'
        else:
            log.info(f"Invalid file type {file_format}. Nothing to process.")
            return

        try:
            database.execute(sql_source)

            give_up = False
            while not give_up:
                try:
                    self.client.head_object(Bucket=self.bucket_name, Key=f"{fldr}/{filenm}")
                    give_up = True
                except botocore.exceptions.ClientError as e:
                    if e.response['Error']['Code'] == "404":
                        log.warning(f"{filenm} does not exist in {self.bucket_name}/{fldr}. Waiting... {counter}")
                        time.sleep(5)
                if counter > 12:
                    log.error(
                        f"{filenm} does not exist in {self.bucket_name}/{fldr}. "
                        f"This file has to be 'renamed' as {filename}. "
                        f"Already waited for a minute. Please check. Exiting."
                    )
                    give_up = True
                    exit(99)
                else:
                    counter += 1

            self.client.copy_object(Bucket=f"{self.bucket_name}", CopySource=f"{self.bucket_name}/{fldr}/{filenm}", Key=f"{filename}")
            self.client.delete_object(Bucket=f"{self.bucket_name}", Key=f"{fldr}/{filenm}")

        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                log.error("The object does not exist.")
                raise e
            else:
                raise e

    def load_into_table_from_s3(
            self,
            database: DatabaseMetadata,
            table_name: str = None,
            filename: str = None,
            delimiter: str = None,
            region: str = None,
        ):
        log = logging.getLogger(__name__)

        if region is None:
            region = self.region_name
            if region is None:
                region = 'us-east-1'

        password = self.get_password()

        sql_source = f"""
            copy {table_name} from 's3://{self.bucket_name}/{filename}' 
            credentials  'aws_access_key_id={self.user_id};aws_secret_access_key={password}' 
            DELIMITER '{delimiter}' CSV NULL '' IGNOREHEADER 1
            REGION '{region}'
            ;
            commit;
            """

        try:
            database.execute(sql_source)

            log.info(f"Loaded {self.bucket_name}/{filename} into {table_name}.")

        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                log.error("The object does not exist.")
                raise e
            else:
                raise e



import warnings
from time import sleep

from config_wrangler.config_templates.credentials import Credentials

import bi_etl.config.notifiers_config as notifiers_config
from bi_etl.notifiers.notifier_base import NotifierBase


class Slack(NotifierBase):
    def __init__(self, config_section: notifiers_config.SlackNotifier):
        super().__init__()
        self.config_section = config_section

        try:
            from slack_sdk import WebClient
            self.log.debug("Using slack_sdk import")
            self._client_version = 3
            from slack_sdk.errors import SlackApiError
        except ImportError:
            # noinspection PyUnresolvedReferences
            from slack import WebClient
            self.log.debug("Using WebClient v2+ import")
            self._client_version = 2
            from slack.errors import SlackApiError
        self.SlackApiError = SlackApiError

        # We need to support putting the token in keepass/keyring
        # slack_token = self.config_section.token
        if self.config_section.token is None or self.config_section.token == '':
            slack_token_credentials = Credentials(
                user_id='slack',
                password_source=self.config_section.token_source,
                raw_password=self.config_section.token,
                keyring_section=self.config_section.keyring_section,
                # keepass_config=self.config_section.keepass_config,
                keepass=self.config_section.keepass,
                keepass_group=self.config_section.keepass_group,
                keepass_title=self.config_section.keepass_title,
            )
            config_section.set_as_child('slack_token_credentials', slack_token_credentials)
            slack_token = slack_token_credentials.get_password()
        else:
            slack_token = self.config_section.token

        self.slack_client = WebClient(slack_token)
        self.slack_channel = self.config_section.channel
        self.mention = self.config_section.mention
        self._status_channel = None
        self._status_ts = None

        if self.slack_channel is None or self.slack_channel == 'OVERRIDE_THIS_SETTING':
            self.log.warning("Slack channel not set. No slack messages will be sent.")
            self.slack_channel = None

    def _post_message(self, text: str, link_names: bool = False):
        retry = True
        result_data = None
        while retry:
            # https://api.slack.com/methods/chat.postMessage
            try:
                result = self.slack_client.chat_postMessage(
                    channel=self.slack_channel,
                    text=text,
                    link_names=link_names,
                )
                retry = False
                try:
                    result_data = result.data
                    if result_data is None:
                        self.log.error(f"API result has result.data of None {result}")
                except (AttributeError, TypeError):
                    result_data = None
                    self.log.error(f"API result did not include data: {result}")

            except self.SlackApiError as e:
                self.log.error(e)
                if e.response['error'] == 'ratelimited':
                    self.log.info('Waiting for slack ratelimited to clear')
                    sleep(1.5)
                    # See retry loop above
                else:
                    raise
        return result_data

    def _post_update(
            self,
            channel: str,  # Note this does not accept the friendly channel names (see send results)
            ts: str,
            text: str,
            link_names: bool = False
    ):
        # Updates are considered non-critical so we don't retry
        try:
            self.slack_client.chat_update(
                channel=channel,
                ts=ts,
                text=text,
                link_names=link_names,
            )
        except self.SlackApiError as e:
            self.log.error(e)
            if e.response['error'] == 'ratelimited':
                # Updates are considered non-critical so rate limiting means we skip the update
                pass
            else:
                raise

    def send(self, subject, message, sensitive_message=None, attachment=None, throw_exception=False):
        # Clear the status timestamp so that we create a new status message below this non-status message
        self._status_ts = None
        if self.slack_channel is not None and self.slack_channel != '':
            if subject and message:
                message_to_send = f"{subject}: {message}"
            else:
                if message:
                    message_to_send = message
                else:
                    message_to_send = subject

            if self.mention:
                message_to_send += ' ' + self.mention
                link_names = True
            else:
                link_names = False

            self._post_message(text=message_to_send, link_names=link_names)

        else:
            self.log.info("Slack message not sent: {}".format(message))

    def post_status(self, status_message):
        """
        Send a temporary status messages that gets overwritten with the next status message that is sent.

        Parameters
        ----------
        status_message

        Returns
        -------

        """
        if self._client_version == 1:
            warnings.warn(f"Slack client v1 not supported for post_status")
            return None

        if self.slack_channel is None or self.slack_channel == '':
            self.log.warning(f"slack_channel = '{self.slack_channel}'")
            return None

        if self._status_ts is None:
            result_data = self._post_message(text=status_message)
            if result_data is not None:
                if 'ts' in result_data and 'channel' in result_data:
                    self._status_ts = result_data['ts']
                    self._status_channel = result_data['channel']
        else:
            self._post_update(
                channel=self._status_channel,
                ts=self._status_ts,
                text=status_message
            )



"""
Created on Sep 17, 2014

@author: Derek Wood
"""
# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import functools
import typing
from copy import copy
from datetime import datetime, date
from operator import attrgetter
from typing import Iterable

import sqlalchemy
from sqlalchemy.sql import sqltypes, functions
from sqlalchemy.sql.expression import text
from sqlalchemy.sql.schema import Column

from bi_etl.components.etlcomponent import ETLComponent
from bi_etl.components.row.row import Row
from bi_etl.components.row.row_iteration_header import RowIterationHeader
from bi_etl.database import DatabaseMetadata
from bi_etl.exceptions import NoResultFound, MultipleResultsFound
from bi_etl.lookups.lookup import Lookup
from bi_etl.scheduler.task import ETLTask
from bi_etl.statistics import Statistics
from bi_etl.utility import dict_to_str

__all__ = ['ReadOnlyTable']


# Pylint does not like references to self.table.columns aka self.columns
# pylint: disable=unsupported-membership-test, unsubscriptable-object, not-an-iterable


class ReadOnlyTable(ETLComponent):
    """ 
    Reads all columns from a database table or view. 
    Rows can be filtered using the :py:meth:`~bi_etl.components.readonlytable.ReadOnlyTable.where` method.
    
    Parameters
    ----------
    task : ETLTask
        The  instance to register in (if not None)

    database : bi_etl.scheduler.task.Database
        The database to find the table/view in.

    table_name : str
        The name of the table/view.

    table_name_case_sensitive: bool
        Should the table name be treated in a case-sensitive manner?

        If false, it will convert the table name to lower case which indicates to SQLAlchemy that it should be
        not case-sensitive in the Oracle dialect.

        if None, this defaults based on the case sensitivity of the SQLAlchemy dialect being used.

        NOTE from SQLAlchemy:
            In Oracle, the data dictionary represents all case-insensitive identifier names using UPPERCASE text.
            SQLAlchemy on the other hand considers an all-lower case identifier name to be case insensitive.
            The Oracle dialect converts all case insensitive identifiers to and from those two formats during
            schema level communication, such as reflection of tables and indexes.
            Using an UPPERCASE name on the SQLAlchemy side indicates a case sensitive identifier,
            and SQLAlchemy will quote the name - this will cause mismatches against data dictionary data
            received from Oracle, so unless identifier names have been truly created as case sensitive
            (i.e. using quoted names), all lowercase names should be used on the SQLAlchemy side.
            https://docs.sqlalchemy.org/en/14/dialects/oracle.html?highlight=oracle#module-sqlalchemy.dialects.oracle.cx_oracle


    include_only_columns : list, optional
        Optional. A list of specific columns to include when reading the table/view.
        All other columns are excluded.

    exclude_columns :
        Optional. A list of columns to exclude when reading the table/view.
         
    Attributes
    ----------    
    delete_flag : str, optional
        The name of the delete_flag column, if any.
    
    delete_flag_yes : str, optional
        The value of delete_flag for deleted rows.
    
    delete_flag_no : str, optional
        The value of delete_flag for *not* deleted rows.
        
    special_values_descriptive_columns: list, optional
         A list of columns that should get longer descriptive text (e.g. 'Missing' instead of '?') in 
         :meth:`get_missing_row`, 
         :meth:`get_invalid_row`, 
         :meth:`get_not_applicable_row`, 
         :meth:`get_various_row`
    
    log_first_row : boolean
        Should we log progress on the the first row read. *Only applies if Table is used as a source.*
        (inherited from ETLComponent)
        
    max_rows : int, optional
        The maximum number of rows to read. *Only applies if Table is used as a source.*
        (inherited from ETLComponent)
        
    maintain_cache_during_load: boolean
        Default = True. Should we maintain the lookup caches as we load records.
        Can safely be set to False for sources that will never use a key combination twice
        during a single load. Setting it to False should improve performance.
        
    primary_key: list
        The name of the primary key column(s). Only impacts trace messages.  Default=None.
        If not passed in, will use the database value, if any.
        (inherited from ETLComponent)

    natural_key: list
        The list of natural key columns (as Column objects).
        Default is None
        
    progress_frequency: int
        How often (in seconds) to output progress messages. None for no progress messages.
        (inherited from ETLComponent)
    
    progress_message: str
        The progress message to print. Default is ``"{logical_name} row # {row_number}"``. 
        Note ``logical_name`` and ``row_number`` subs.
        (inherited from ETLComponent)
    
    """
    PK_LOOKUP = 'PK'
    NK_LOOKUP = 'NK'

    def __init__(self,
                 task: typing.Optional[ETLTask],
                 database: DatabaseMetadata,
                 table_name: str,
                 table_name_case_sensitive: bool = None,
                 schema: str = None,
                 exclude_columns: set = None,
                 include_only_columns: set = None,
                 **kwargs
                 ):
        # Don't pass kwargs up. They should be set here at the end
        super(ReadOnlyTable, self).__init__(task=task,
                                            logical_name=table_name,
                                            )
        self.trace_sql = False
        self.always_fallback_to_db = True
        self._maintain_cache_during_load = True

        self.__compile_cache = {}
        self.__delete_flag = None
        self.delete_flag_yes = 'Y'
        self.delete_flag_no = 'N'

        self.special_values_descriptive_columns = set()
        self.special_values_descriptive_min_length = 14  # Long enough to hold 'Not Applicable'
        self._special_row_header = None

        self.database = database
        self._table = None
        if table_name_case_sensitive is None:
            table_name_case_sensitive = self.database.dialect.case_sensitive
        self._table_name_case_sensitive = table_name_case_sensitive
        self._columns = None
        self._column_names = None
        self._column_name_index = None
        self._excluded_columns = None
        self._table_name = table_name
        self.schema = schema
        if table_name is not None:
            # Note from sqlalchemy:
            # Names which contain no upper case characters will be treated as case insensitive names, and will not 
            # be quoted unless they are a reserved word. 
            # Names with any number of upper case characters will be quoted and sent exactly. Note that this behavior
            # applies even for databases which standardize upper case names as case insensitive such as Oracle.
            if '.' in table_name:
                self.schema, table_name = table_name.split('.', 1)

            if not table_name_case_sensitive:
                table_name = table_name.lower()
            try:
                self.table = sqlalchemy.schema.Table(table_name, database, schema=self.schema, autoload=True, quote=False)
            except Exception as e:
                try:
                    self.log.debug(f"Exception {repr(e)} occurred while obtaining definition of {table_name} from the database schema {self.schema}")
                finally:
                    raise

        self.__natural_key_override = False
        self.__natural_key = None

        if exclude_columns:
            self.exclude_columns(exclude_columns)

        if include_only_columns:
            self.include_only_columns(include_only_columns)

        self.custom_special_values = dict()

        self.auto_commit = True
        self._connections_used = set()

        # Should be the last call of every init
        self.set_kwattrs(**kwargs)

    def __reduce_ex__(self, protocol):
        dict_to_pass = copy(self.__dict__)
        del dict_to_pass['_table']

        return (
            # A callable object that will be called to create the initial version of the object.
            self.__class__,

            # A tuple of arguments for the callable object. An empty tuple must be given if the callable does not accept any argument
            (
                self.task,
                self.database,
                self._table_name,
                self._table_name_case_sensitive,
                self.schema,
                self._excluded_columns,
            ),

            # Optionally, the object’s state, which will be passed to the object’s __setstate__() method as previously described.
            # If the object has no such method then, the value must be a dictionary and it will be added to the object’s __dict__ attribute.
            dict_to_pass,

            # Optionally, an iterator (and not a sequence) yielding successive items.
            # These items will be appended to the object either using obj.append(item) or, in batch, using obj.extend(list_of_items).

            # Optionally, an iterator (not a sequence) yielding successive key-value pairs.
            # These items will be stored to the object using obj[key] = value

            # PROTOCOL 5+ only
            # Optionally, a callable with a (obj, state) signature.
            # This callable allows the user to programmatically control the state-updating behavior of a specific object,
            # instead of using obj’s static __setstate__() method.
            # If not None, this callable will have priority over obj’s __setstate__().
        )

    def __repr__(self):
        template = "{cls}(task={task}," \
            "logical_name={logical_name}," \
            "primary_key={primary_key}," \
            "delete_flag={delete_flag})"
        return template.format(
            cls=self.__class__.__name__,
            task=self.task,
            logical_name=self.logical_name,
            primary_key=self.primary_key,
            delete_flag=self.delete_flag,
        )

    @property
    def delete_flag(self):
        return self.__delete_flag

    @delete_flag.setter
    def delete_flag(self, value):
        self.__delete_flag = value
        if self.__delete_flag is not None:
            self.custom_special_values[self.delete_flag] = self.delete_flag_no

    @property
    def table_name(self):
        """
        The table name
        """
        return self._table_name

    @property
    def qualified_table_name(self):
        """
        The table name
        """
        if self._table.schema is not None:
            return self._table.schema + '.' + self._table_name
        else:
            return self._table_name

    @property
    def quoted_qualified_table_name(self):
        """
        The table name
        """
        if self._table.schema is not None:
            return f'"{self._table.schema}"."{self._table_name}"'
        else:
            return f'"{self._table_name}"'

    @property
    def table(self):
        return self._table

    @table.setter
    def table(self, new_table_object):
        self._table = new_table_object
        self._columns = list(self._table.columns)
        self.column_names = list(map(attrgetter('name'), self._columns))
        self._column_name_index = dict()
        for column in self._columns:
            index_version_of_column_name = column.name.lower()
            if index_version_of_column_name in self._column_name_index:
                existing_entry = self._column_name_index[index_version_of_column_name]
                if isinstance(existing_entry, list):
                    existing_entry.append(column)
                else:
                    new_list = list()
                    new_list.append(existing_entry)
                    new_list.append(column)
                    self._column_name_index[index_version_of_column_name] = new_list
                self._column_name_index[index_version_of_column_name] = column
            else:  # New entry
                self._column_name_index[index_version_of_column_name] = column
        self.logical_name = self._table_name = new_table_object.name
        # Get the primary key from the table
        self.primary_key = list(self._table.primary_key)

    @property
    def maintain_cache_during_load(self) -> bool:
        return self._maintain_cache_during_load

    @maintain_cache_during_load.setter
    def maintain_cache_during_load(self, value: bool):
        self._maintain_cache_during_load = value

    def set_columns(self, columns):
        # Remove columns from the table
        for col in self._columns:
            col.table = None
        # Clear table from columns passed in
        for col in columns:
            col.table = None
        # Remove the table definition so we can add it back below
        self._columns = columns
        self._column_names = None
        self.database.remove(self.table)
        self.table = sqlalchemy.schema.Table(self.table_name, self.database, *self._columns, schema=self.schema, quote=False)

    def exclude_columns(self, columns_to_exclude: set):
        """
        Exclude columns from the table. Removes them from all SQL statements.
        
        columns_to_exclude :
            A list of columns to exclude when reading the table/view.
        """

        # =======================================================================
        # Implementation notes:
        # This method accesses protected _columns in the table and removes them.
        # 
        # This implementation needs to reverse what this does...
        #     # see https://github.com/zzzeek/sqlalchemy/blob/master/lib/sqlalchemy/sql/selectable.py
        #     # see class TableClause(Immutable, FromClause):
        #     def append_column(self, c):
        #         self._columns[c.key] = c
        #         c.table = self
        # ===================================================================
        # for ex_name in columns_to_exclude:                
        #     if ex_name in self.columns:
        #         exclude_column_obj = self.columns[ex_name]
        #         self.log.debug('Excluding column {}'.format(exclude_column_obj))
        #         exclude_column_obj.table = None
        #         self.table._columns.remove(exclude_column_obj)
        # self._columns = list(self.table.columns)
        # self.column_names = list(map(attrgetter('name'), self._columns))
        # =======================================================================
        if columns_to_exclude is not None:
            # This method builds a new Table object with the non-excluded columns.
            if self._excluded_columns is None:
                self._excluded_columns = frozenset(columns_to_exclude)
            else:
                self._excluded_columns = self._excluded_columns | columns_to_exclude

        # noinspection PyTypeChecker
        for ex_name in self._excluded_columns:
            try:
                exclude_column_obj = self.get_column(ex_name)
                if exclude_column_obj in self._columns:
                    self.log.debug('Excluding column {}'.format(exclude_column_obj))
                    self._columns.remove(exclude_column_obj)
            except KeyError:
                # Already not there
                pass
        self.set_columns(self._columns)

    def include_only_columns(self, columns_to_include: set):
        """
        Include only specified columns in the table definition.
        Columns that are non included are removed them from all SQL statements.

        columns_to_include : list
            A list of columns to include when reading the table/view.
        """
        columns_to_exclude = self.column_names_set.difference(columns_to_include)
        self.exclude_columns(set(columns_to_exclude))

    def is_connected(self, connection_name: typing.Optional[str] = None) -> bool:
        return self.database.is_connected(connection_name)

    def connection(
            self,
            connection_name: typing.Optional[str] = None,
            open_if_not_exist: bool = True,
            open_if_closed: bool = True,
            auto_commit: bool = None,
    ) -> sqlalchemy.engine.base.Connection:
        connection_name = self.database.resolve_connection_name(connection_name)
        self._connections_used.add(connection_name)
        if auto_commit is None:
            auto_commit = self.auto_commit
        return self.database.connection(
            connection_name=connection_name,
            open_if_not_exist=open_if_not_exist,
            open_if_closed=open_if_closed,
            auto_commit=auto_commit,
        )

    def close_connection(self, connection_name: str = None):
        connection_name = self.database.resolve_connection_name(connection_name)
        self._connections_used.remove(connection_name)
        self.database.close_connection()

    def close_connections(self, exceptions: typing.Optional[set] = None):
        self.database.close_connections(exceptions=exceptions)

    def close(self, error: bool = False):
        if not self.is_closed:
            for lookup in self.lookups.values():
                lookup.add_size_to_stats()
            if len(self.__compile_cache) > 0:
                cache_stats = self.get_stats_entry('compile cache', print_start_stop_times=False)
                cache_stats['entries'] = len(self.__compile_cache)
                del self.__compile_cache
            self.close_connections()
            super().close(error=error)

    def __del__(self):
        # noinspection PyBroadException
        try:
            self.close_connections()
        except Exception:
            pass

    def __exit__(self, exit_type, exit_value, exit_traceback):  # @ReservedAssignment
        # Close the database connection
        self.close()

    def execute(
            self,
            statement,
            *list_params,
            connection_name: str = None,
            auto_commit: bool = None,
            **params
    ) -> sqlalchemy.engine.ResultProxy:
        """

        Parameters
        ----------
        statement:
            The SQL statement to execute

        connection_name:
            Name of the pooled connection to use
            Defaults to DEFAULT_CONNECTION_NAME

        Returns
        -------
        sqlalchemy.engine.ResultProxy with results
        """
        # compiled_cache created huge memory usage. It seems like each lookup created it's own entry
        # execution_options(compiled_cache=self.__compile_cache)
        if self.trace_sql:
            self.log.debug(f'SQL={statement}')
            try:
                self.log.debug(f'parameters={dict_to_str(statement.parameters)}')
            # pylint: disable=broad-except
            except Exception as e:
                self.log.debug(e)
            self.log.debug(f'list_params={list_params}')
            self.log.debug(f'params={params}')
            self.log.debug('-------------------------')
        connection_name = self.database.resolve_connection_name(connection_name)
        self._connections_used.add(connection_name)
        if auto_commit is None:
            auto_commit = self.auto_commit
        return self.database.execute(
            statement,
            *list_params,
            connection_name=connection_name,
            auto_commit=auto_commit,
            transaction=False,
            auto_close=False,
            **params
        )

    def get_one(self, statement=None, connection_name: str = 'get_one'):
        """
        Executes and gets one row from the statement.

        Parameters
        ----------
        statement:
            The SQL statement to execute

        connection_name:
            Name of the pooled connection to use
            Defaults to 'get_one'
        
        Returns
        -------
        row : :class:`~bi_etl.components.row.row_case_insensitive.Row`
            The row returned
            
        Raises
        ------
        NoResultFound
            No rows returned.
            
        MultipleResultsFound
            More than one row was returned.
        """
        if statement is None:
            statement = self.select()
        results = self.execute(
            statement,
            connection_name=connection_name,
            # Auto commit so that this SELECT does not hold a lock
            auto_commit=True,
        )
        row1 = results.fetchone()
        if row1 is None:
            raise NoResultFound()
        row2 = results.fetchone()
        if connection_name == 'get_one':
            self.close_connection(connection_name)
        if row2 is not None:
            raise MultipleResultsFound([row1, row2])
        return self.Row(row1)

    def select(
            self,
            column_list: typing.Optional[list] = None,
            exclude_cols: typing.Optional[frozenset] = None
    ) -> sqlalchemy.sql.expression.Select:
        """
        Builds a select statement for this table. 
        
        Returns
        -------
        statement: 
                    
        """
        if exclude_cols is not None and column_list is not None:
            raise ValueError("select can't accept both column_list and exclude_cols")
        if column_list is None:
            if exclude_cols is None:
                return sqlalchemy.select(self.columns)
            else:
                # Make sure all entries are column objects
                exclude_col_objs = frozenset([self.get_column(c) for c in exclude_cols])
                filtered_column_list = [c for c in self.columns if c not in exclude_col_objs]
                return sqlalchemy.select(filtered_column_list)
        else:
            column_obj_list = list()
            # noinspection PyTypeChecker
            for c in column_list:
                if isinstance(c, str):
                    column_obj_list.append(self.get_column(c))
                else:
                    column_obj_list.append(c)
            return sqlalchemy.select(column_obj_list).execution_options(autocommit=True)

    def _generate_key_values_dict(self, key_names=None, key_values=None, lookup_name=None, other_values_dict=None):

        if key_names is not None and lookup_name is not None:
            raise ValueError('Both key_names and lookup_name provided. Please use one or the other')

        # Get key names from key_names or lookup
        if key_names is None:
            if lookup_name is not None:
                key_names = self.get_lookup(lookup_name).lookup_keys
            else:
                self._check_pk_lookup()
                lookup_name = self._get_pk_lookup_name()
                key_names = self.primary_key
        else:
            # Handle case where we have a single key name item and not a list or dict
            if isinstance(key_names, str):
                key_names = [key_names]

        # Handle case where we have a single key value item and not a list
        if isinstance(key_values, str):
            key_values = [key_values]

        key_values_dict = dict()
        if isinstance(key_values, typing.MutableMapping):
            for key_name in key_names:
                if key_name in key_values:
                    key_values_dict[key_name] = key_values[key_name]
                elif other_values_dict is not None and key_name in other_values_dict:
                    key_values_dict[key_name] = other_values_dict[key_name]
                else:
                    raise ValueError('No key value provided for {}'.format(key_name))
        elif key_values is None:
            if other_values_dict is None:
                raise ValueError('No key values provided')
            else:
                for key_name in key_names:
                    if key_name in other_values_dict:
                        key_values_dict[key_name] = other_values_dict[key_name]
                    else:
                        raise ValueError('No key value provided for {}'.format(key_name))
        else:
            # Otherwise we'll assume key_values is a list or iterable like one
            key_values_dict = dict()
            assert len(key_values) == len(key_names), 'key values list does not match length of key names list'
            for key_name, key_value in zip(key_names, key_values):
                key_values_dict[key_name] = key_value
        return key_values_dict, lookup_name

    def cache_iterator(self):
        pk_lookup = None
        try:
            pk_lookup = self.get_pk_lookup()
            if not pk_lookup.cache_enabled:
                for lookup_name in self.__lookups:
                    lookup = self.get_lookup(lookup_name)
                    if lookup.cache_enabled:
                        pk_lookup = lookup
                        break
        except KeyError:
            pass

        if pk_lookup is None:
            raise LookupError("Unable to find filled lookup for cache_iterator")
        else:
            # Note in this case the outer call where / iter_result will process the criteria
            self.log.debug(f"Using lookup {pk_lookup} as source for cache_iterator")
            return pk_lookup

    # We add the following args at this level
    #   criteria: so they can be passed down to the database
    #   use_cache_as_source: to we can use the cache instead of the database
    #   stats_id, parent_stats: So we can capture SQL execution time in the right place.
    def _raw_rows(self,
                  criteria_list: list = None,
                  criteria_dict: dict = None,
                  order_by: list = None,
                  column_list: list = None,
                  exclude_cols: frozenset = None,
                  use_cache_as_source: bool = None,
                  connection_name: str = None,
                  stats_id: str = None,
                  parent_stats: Statistics = None):
        """
        Iterate over rows matching ``criteria``
        
        Parameters
        ----------
        criteria_list : string or list of strings
            Each string value will be passed to :meth:`sqlalchemy.sql.expression.Select.where`.
            http://docs.sqlalchemy.org/en/rel_1_0/core/selectable.html?highlight=where#sqlalchemy.sql.expression.Select.where
        criteria_dict : dict
            Dict keys should be columns, values are set using = or in
        order_by: string or list of strings
            Each value should represent a column to order by.
        exclude_cols:
            List of columns to exclude from the results. (Only if getting from the database)
        use_cache_as_source: bool
            Should we read rows from the cache instead of the table
        stats_id: string
            Name of this step for the ETLTask statistics.
        parent_stats: bi_etl.statistics.Statistics
            Optional Statistics object to nest this steps statistics in.
            Default is to place statistics in the ETLTask level statistics.
        connection_name:
            The name of the pooled connection to use
    
        Yields
        ------
        row: :class:`~bi_etl.components.row.row_case_insensitive.Row`
            Row object with contents of a table/view row that matches ``criteria_list`` and ``criteria_dict``
        """
        if stats_id is None:
            stats_id = self.default_stats_id
        stats = self.get_stats_entry(stats_id=stats_id, parent_stats=parent_stats)

        if use_cache_as_source is None:
            use_cache_as_source = True
            use_cache_as_source_requested = False
        else:
            use_cache_as_source_requested = use_cache_as_source

        pk_lookup = None
        if use_cache_as_source:
            if not (self.cache_filled and self.cache_clean):
                use_cache_as_source = False
                if use_cache_as_source_requested:
                    self.log.warning("Cache not filled requires using database as source for {}".format(stats))
            elif order_by is not None:
                use_cache_as_source = False
                if use_cache_as_source_requested:
                    self.log.warning("where had to use DB source to honor order_by (and possibly criteria)")
            elif criteria_list is not None:
                if use_cache_as_source_requested:
                    raise ValueError(f"Non dict criteria ({criteria_list}) requires using database as source for {stats} on {self}")
            else:
                # Find the filled cache
                try:
                    pk_lookup = self.get_pk_lookup()
                    if not pk_lookup.cache_enabled:
                        use_cache_as_source = False
                        if use_cache_as_source_requested:
                            self.log.debug("PK cache not filled. "
                                           "Looking for another lookup to use for {}".format(stats))
                        for lookup_name in self.__lookups:
                            lookup = self.get_lookup(lookup_name)
                            if lookup.cache_enabled:
                                use_cache_as_source = True
                                pk_lookup = lookup
                                break
                        if not use_cache_as_source:
                            if use_cache_as_source_requested:
                                self.log.debug("Unable to find filled lookup. "
                                               "Requires using database as source for {}".format(stats))
                except KeyError:
                    use_cache_as_source = False
                    if use_cache_as_source_requested:
                        self.log.debug("KeyError finding lookup. "
                                       "Requires using database as source for {}".format(stats))

        if use_cache_as_source:
            # Note in this case the outer call where / iter_result will process the criteria
            self.log.debug("Using lookup {} as source for {}".format(pk_lookup, stats))
            self._iterator_applied_filters = False
            return pk_lookup
        else:
            self.log.debug("Using database as source for {}".format(stats))
            self._iterator_applied_filters = True
            stmt = self.select(column_list=column_list,
                               exclude_cols=exclude_cols,
                               )
            if criteria_dict is not None:
                for col, value in criteria_dict.items():
                    stmt = stmt.where(self.get_column(col) == value)
            if criteria_list is not None:
                if isinstance(criteria_list, list):
                    for c in criteria_list:
                        if isinstance(c, str):
                            stmt = stmt.where(text(c))
                        elif isinstance(c, typing.Mapping):
                            for col, value in c.items():
                                stmt = stmt.where(self.get_column(col) == value)
                        else:
                            stmt = stmt.where(c)
                elif isinstance(criteria_list, str):
                    stmt = stmt.where(text(criteria_list))
                else:
                    stmt = stmt.where(criteria_list)
            if order_by is not None:
                if isinstance(order_by, list):
                    stmt = stmt.order_by(*order_by)
                else:
                    stmt = stmt.order_by(order_by)
            self.log.debug(f'Table Read SQL=\n{stmt}\n---End Fill cache SQL')
            stats.timer.start()
            select_result = self.execute(
                stmt,
                connection_name=connection_name,
                # Auto commit so that this SELECT does not hold a lock
                auto_commit=True,
            )
            stats.timer.stop()
            return select_result

    def where(self,
              criteria_list: list = None,
              criteria_dict: dict = None,
              order_by: list = None,
              column_list: typing.List[typing.Union[Column, str]] = None,
              exclude_cols: typing.FrozenSet[typing.Union[Column, str]] = None,
              use_cache_as_source: bool = None,
              connection_name: str = 'select',
              progress_frequency: int = None,
              stats_id: str = None,
              parent_stats: Statistics = None,
              ) -> Iterable[Row]:
        """

        Parameters
        ----------
        criteria_list:
            Each string value will be passed to :meth:`sqlalchemy.sql.expression.Select.where`.
            http://docs.sqlalchemy.org/en/rel_1_0/core/selectable.html?highlight=where#sqlalchemy.sql.expression.Select.where
        criteria_dict:
            Dict keys should be columns, values are set using = or in
        order_by:
            List of sort keys
        column_list:
            List of columns (str or Column)
        exclude_cols
        use_cache_as_source
        connection_name:
            Name of the pooled connection to use
        progress_frequency
        stats_id
        parent_stats

        Returns
        -------
        rows

        """
        result_rows_iter = self._raw_rows(criteria_list=criteria_list,
                                          criteria_dict=criteria_dict,
                                          order_by=order_by,
                                          column_list=column_list,
                                          exclude_cols=exclude_cols,
                                          use_cache_as_source=use_cache_as_source,
                                          connection_name=connection_name,
                                          stats_id=stats_id,
                                          parent_stats=parent_stats,
                                          )
        return self.iter_result(result_rows_iter,
                                criteria_dict=criteria_dict,
                                progress_frequency=progress_frequency,
                                stats_id=stats_id,
                                parent_stats=parent_stats,
                                columns_in_order=column_list,
                                )

    def order_by(self,
                 order_by: list,
                 stats_id: str = None,
                 parent_stats: Statistics = None,
                 ) -> Iterable[Row]:
        """
        Iterate over all rows in order provided.
        
        Parameters
        ----------
        order_by: string or list of strings
            Each value should represent a column to order by.
        stats_id: string
            Name of this step for the ETLTask statistics.
        parent_stats: bi_etl.statistics.Statistics
            Optional Statistics object to nest this steps statistics in.
            Default is to place statistics in the ETLTask level statistics.          
    
        Yields
        ------
        row: :class:`~bi_etl.components.row.row_case_insensitive.Row`
            :class:`~bi_etl.components.row.row_case_insensitive.Row` object with contents of a table/view row
        """
        return self.where(order_by=order_by,
                          stats_id=stats_id,
                          parent_stats=parent_stats
                          )

    @property
    def columns(self) -> typing.List[Column]:
        """
        A named-based collection of :class:`sqlalchemy.sql.expression.ColumnElement` objects in this table/view. 
        
        """
        return self._columns

    def get_column(self, column: typing.Union[str, Column]) -> Column:
        """
        Get the :class:`sqlalchemy.sql.expression.ColumnElement` object for a given column name.
        """
        if isinstance(column, Column):
            if column.table == self.table:
                return column
            else:
                return self.get_column(column.name)
        else:
            if column.lower() in self._column_name_index:
                index_entry = self._column_name_index[column.lower()]
                if isinstance(index_entry, list):
                    # More than one column with that name. 
                    # Case sensitivity required
                    raise KeyError('{} does not have a column named {} '
                                   'however multiple other case versions exist'.format(self.table_name, column))
                else:
                    return index_entry
            else:
                raise KeyError(
                    '{} does not have a column named {} (or lowercase {}) it does have {}'.format(self.table_name, column, column.lower(), self._column_name_index.keys()))

    def get_column_name(self, column):
        """
        Get the column name given a possible :class:`sqlalchemy.sql.expression.ColumnElement` object.
        """
        if isinstance(column, Column):
            return self.table.columns[column.name].name
        else:
            return self.table.columns[column].name

    def max(self, column, where=None, connection_name: str = 'max'):
        """
        Query the table/view to get the maximum value of a given column. 
        
        Parameters
        ----------
        column: str or :class:`sqlalchemy.sql.expression.ColumnElement`.
            The column to get the max value of

        where: string or list of strings
            Each string value will be passed to :meth:`sqlalchemy.sql.expression.Select.where`
            http://docs.sqlalchemy.org/en/rel_1_0/core/selectable.html?highlight=where#sqlalchemy.sql.expression.Select.where

        connection_name:
            Name of the pooled connection to use
            Defaults to 'max'
        
        Returns
        -------
        max : depends on column datatype
            
        """
        c = self.get_column(column)
        stmt = self.select([functions.max(c).label("max_1")])
        if where is not None:
            if isinstance(where, list):
                for c in where:
                    stmt = stmt.where(c)
            else:
                stmt = stmt.where(where)
        max_row = self.get_one(stmt, connection_name=connection_name)
        if connection_name == 'max':
            self.close_connection(connection_name)
        max_value = max_row['max_1']
        return max_value

    def count(self, column: str = None, where=None) -> int:
        """
        Query the table/view to get the count of a given column.

        Parameters
        ----------
        column: str or :class:`sqlalchemy.sql.expression.ColumnElement`.
            The column to get the max value of
        where: string or list of strings
            Each string value will be passed to :meth:`sqlalchemy.sql.expression.Select.where`
            http://docs.sqlalchemy.org/en/rel_1_0/core/selectable.html?highlight=where#sqlalchemy.sql.expression.Select.where

        Returns
        -------
        count : int

        """
        if column is not None:
            column = self.get_column(column)
        stmt = self.select([functions.count(column).label("count_1")]).select_from(self.table)
        if where is not None:
            if isinstance(where, list):
                for c in where:
                    stmt = stmt.where(c)
            else:
                stmt = stmt.where(where)
        row = self.get_one(stmt)
        count_value = row['count_1']
        return count_value

    @property
    @functools.lru_cache(maxsize=16)
    def row_name(self):
        return str(self.table)

    def get_column_special_value(
            self,
            column: Column,
            short_char: str,
            long_char: str,
            int_value: int,
            date_value: datetime,
            use_custom_special_values: bool = 'Y'
        ) -> object:
        target_type = column.type
        special_value = None
        if use_custom_special_values and self.custom_special_values and column.name in self.custom_special_values:
            custom_value = self.custom_special_values[column.name]
            if custom_value == '[short_char]':
                special_value = short_char
            elif custom_value == '[long_char]':
                special_value = long_char
            elif custom_value == '[int_value]':
                special_value = int_value
            elif custom_value == '[date_value]':
                special_value = date_value
            else:
                # Note: We test for setitem because dict like containers have it but str doesn't have it.
                if hasattr(custom_value, '__setitem__'):
                    if short_char in custom_value:
                        # noinspection PyTypeChecker
                        special_value = custom_value[short_char]
                    else:
                        # This short_char value doesn't have a custom assignment, try again with no custom value
                        special_value = self.get_column_special_value(
                            column=column,
                            short_char=short_char,
                            long_char=long_char,
                            int_value=int_value,
                            date_value=date_value,
                            use_custom_special_values=False
                        )
                else:
                    special_value = custom_value
        elif column.name == self.delete_flag:
            special_value = 'N'
        elif (target_type.python_type == str
              or isinstance(target_type, sqltypes.String)
        ):
            if column.name in self.special_values_descriptive_columns:
                special_value = long_char
            elif column.type.length is None:
                special_value = long_char
            elif (self.special_values_descriptive_min_length
                  and column.type.length >= self.special_values_descriptive_min_length
            ):
                special_value = long_char
            else:
                special_value = short_char
        elif (target_type.python_type in {date, datetime}
              or isinstance(target_type, sqltypes.DATE)
              or isinstance(target_type, sqltypes.DATETIME)
              or isinstance(target_type, sqlalchemy.types.DateTime)
        ):
            special_value = date_value
        elif (target_type.python_type in {int}
              or isinstance(target_type, sqltypes.INTEGER)
              or isinstance(target_type, sqltypes.Numeric)
        ):
            special_value = int_value
        elif (target_type.python_type == float
              or isinstance(target_type, sqltypes.FLOAT)
              or isinstance(target_type, sqltypes.Numeric)
        ):
            special_value = float(int_value)
        return special_value

    def get_special_row(
            self,
            short_char: str,
            long_char: str,
            int_value: int,
            date_value: datetime,
    ):
        if self._special_row_header is None:
            self._special_row_header = self.generate_iteration_header(
                logical_name='get_special_row',
            )
        row = self.row_object(iteration_header=self._special_row_header)
        for column in self.columns:
            row[column.name] = self.get_column_special_value(
                column=column,
                short_char=short_char,
                long_char=long_char,
                int_value=int_value,
                date_value=date_value,
            )
        return row

    def get_missing_row(self):
        """
        Get a :class:`~bi_etl.components.row.row_case_insensitive.Row` 
        with the Missing special values filled in for all columns.
        
        =========== =========
        Type        Value
        =========== =========
        Integer     -9999
        Short Text  '?'
        Long Text   'Missing'
        Date        9999-9-1
        =========== =========
        """
        return self.get_special_row('?', 'Missing', -9999, datetime(9999, 9, 1))

    def get_invalid_row(self):
        """
        Get a :class:`~bi_etl.components.row.row_case_insensitive.Row` 
        with the Invalid special values filled in for all columns.
        
        =========== =========
        Type        Value
        =========== =========
        Integer     -8888
        Short Text  '!'
        Long Text   'Invalid'
        Date        9999-8-1
        =========== ========= 
        
        """
        return self.get_special_row('!', 'Invalid', -8888, datetime(9999, 8, 1))

    def get_not_applicable_row(self):
        """
        Get a :class:`~bi_etl.components.row.row_case_insensitive.Row` 
        with the Not Applicable special values filled in for all columns.
        
        =========== =========
        Type        Value
        =========== =========
        Integer     -7777
        Short Text  '~'
        Long Text   'Not Available'
        Date        9999-7-1
        =========== =========          
        """
        return self.get_special_row('~', 'Not Available', -7777, datetime(9999, 7, 1))

    def get_various_row(self):
        """
        Get a :class:`~bi_etl.components.row.row_case_insensitive.Row` 
        with the Various special values filled in for all columns.
        
        =========== =========
        Type        Value
        =========== =========
        Integer     -6666
        Short Text  '*'
        Long Text   'Various'
        Date        9999-6-1
        =========== =========
        """
        return self.get_special_row('*', 'Various', -6666, datetime(9999, 6, 1))

    def get_none_selected_row(self):
        """
        Get a :class:`~bi_etl.components.row.row_case_insensitive.Row`
        with the None Selected special values filled in for all columns.

        =========== =========
        Type        Value
        =========== =========
        Integer     -5555
        Short Text  '#'
        Long Text   'None Selected'
        Date        9999-5-1
        =========== =========
        """
        return self.get_special_row('#', 'None Selected', -5555, datetime(9999, 5, 1))

    def _check_pk_lookup(self):
        # Check that we have setup the PK lookup.
        # Late binding so that it will take overrides to the default lookup class

        pk_lookup_name = self._get_pk_lookup_name()
        if pk_lookup_name not in self.lookups:
            if self.primary_key:
                self.define_lookup(pk_lookup_name, self.primary_key)

    def fill_cache_from_source(
            self,
            source: ETLComponent,
            progress_frequency: float = 10,
            progress_message="{component} fill_cache current row # {row_number:,}",
            criteria_list: list = None,
            criteria_dict: dict = None,
            column_list: list = None,
            exclude_cols: frozenset = None,
            order_by: list = None,
            assume_lookup_complete: bool = None,
            allow_duplicates_in_src: bool = False,
            row_limit: int = None,
            parent_stats: Statistics = None,
            ):
        """
        Fill all lookup caches from the table.

        Parameters
        ----------
        source:
            Source compontent to get rows from.
        progress_frequency : int, optional
            How often (in seconds) to output progress messages. Default 10. None for no progress messages.
        progress_message : str, optional
            The progress message to print.
            Default is ``"{component} fill_cache current row # {row_number:,}"``. Note ``logical_name`` and ``row_number``
            substitutions applied via :func:`format`.
        criteria_list : string or list of strings
            Each string value will be passed to :meth:`sqlalchemy.sql.expression.Select.where`.
            https://goo.gl/JlY9us
        criteria_dict : dict
            Dict keys should be columns, values are set using = or in
        column_list:
            List of columns to include
        exclude_cols:
            Optional. Columns to exclude when filling the cache
        order_by: list
            list of columns to sort by when filling the cache (helps range caches)
        assume_lookup_complete: boolean
            Should later lookup calls assume the cache is complete?
            If so, lookups will raise an Exception if a key combination is not found.
            Default to False if filtering criteria was used, otherwise defaults to True.
        allow_duplicates_in_src:
            Should we quietly let the source provide multiple rows with the same key values? Default = False
        row_limit: int
            limit on number of rows to cache.
        parent_stats: bi_etl.statistics.Statistics
            Optional Statistics object to nest this steps statistics in.
            Default is to place statistics in the ETLTask level statistics.
        """
        super().fill_cache_from_source(
            source=source,
            progress_frequency=progress_frequency,
            progress_message=progress_message,
            criteria_list=criteria_list,
            criteria_dict=criteria_dict,
            column_list=column_list,
            exclude_cols=exclude_cols,
            order_by=order_by,
            assume_lookup_complete=assume_lookup_complete,
            allow_duplicates_in_src=allow_duplicates_in_src,
            row_limit=row_limit,
            parent_stats=parent_stats,
            )
        # Set the table always_fallback_to_db value based on if criteria
        # were used to load the cache and the assume_lookup_complete parameter
        if criteria_list is not None or criteria_dict is not None:
            # If criteria is used we'll default to not assuming the lookup is complete
            if assume_lookup_complete is None:
                assume_lookup_complete = False
        else:
            # If criteria is NOT used we'll default to assuming the lookup is complete
            if assume_lookup_complete is None:
                assume_lookup_complete = True
        self.always_fallback_to_db = not assume_lookup_complete
        self.log.info(f'Lookups will always_fallback_to_db = {self.always_fallback_to_db}')

    @functools.lru_cache(maxsize=10)
    def get_lookup(self, lookup_name: typing.Optional[str]) -> Lookup:
        if lookup_name is None:
            lookup_name = self._get_pk_lookup_name()
        return super().get_lookup(lookup_name=lookup_name)

    def _get_pk_lookup_name(self):
        return self.get_qualified_lookup_name(self.PK_LOOKUP)

    def get_pk_lookup(self) -> Lookup:
        self._check_pk_lookup()
        return self.get_lookup(self._get_pk_lookup_name())

    def get_primary_key_value_list(self, row) -> list:
        return self.get_pk_lookup().get_list_of_lookup_column_values(row)

    def get_primary_key_value_tuple(self, row) -> tuple:
        return self.get_pk_lookup().get_hashable_combined_key(row)

    @property
    def natural_key(self) -> list:
        """
        Get this tables natural key
        """
        return self.__natural_key

    @natural_key.setter
    def natural_key(self, value: list):
        self.__natural_key_override = True
        nk = list()
        for colname in value:
            k = self.get_column_name(colname)
            nk.append(k)
        self.__natural_key = nk
        self.ensure_nk_lookup()

    def _get_nk_lookup_name(self):
        if self.__natural_key_override:
            return self.get_qualified_lookup_name(self.NK_LOOKUP)
        else:
            return self._get_pk_lookup_name()

    def get_nk_lookup_name(self):
        self.ensure_nk_lookup()
        return self._get_nk_lookup_name()

    def ensure_nk_lookup(self):
        nk_lookup_name = self._get_nk_lookup_name()
        if nk_lookup_name not in self.lookups:
            if self.natural_key:
                self.define_lookup(nk_lookup_name, self.natural_key)

    def get_nk_lookup(self) -> Lookup:
        self.ensure_nk_lookup()
        nk_lookup_name = self._get_nk_lookup_name()
        return self.get_lookup(nk_lookup_name)

    @functools.lru_cache(maxsize=10)
    def get_default_lookup(self, row_iteration_header: RowIterationHeader) -> Lookup:
        pk_lookup = self.get_pk_lookup()
        row_has_pk = pk_lookup.row_iteration_header_has_lookup_keys(row_iteration_header)
        if self.natural_key is None or len(self.natural_key) == 0:
            if row_has_pk:
                return pk_lookup
            else:
                raise LookupError(f"{row_iteration_header} has no keys for {self} PK lookup {pk_lookup} and no NK exists")
        else:
            if row_has_pk:
                return pk_lookup
            else:
                nk_lookup = self.get_nk_lookup()
                if nk_lookup.row_iteration_header_has_lookup_keys(row_iteration_header):
                    return nk_lookup
                else:
                    raise LookupError(f"{row_iteration_header} has no keys for {self} PK {pk_lookup} or NK {nk_lookup}")

    def get_natural_key_value_list(self, row: Row) -> list:
        if self.natural_key is None:
            return self.get_primary_key_value_list(row)
        else:
            natural_key_values = [row[k] for k in self.natural_key]
            return natural_key_values

    def get_natural_key_tuple(self, row) -> tuple:
        if self.natural_key is None:
            return self.get_primary_key_value_tuple(row)
        else:
            return tuple(self.get_nk_lookup().get_hashable_combined_key(row))
            # # We need to make sure to rstrip any string values to that it matches what Lookup uses
            # return tuple([Lookup.rstrip_key_value(row[k]) for k in self.natural_key])

    def get_by_key(self,
                   source_row: Row,
                   stats_id: str = 'get_by_key',
                   parent_stats: Statistics = None, ) -> Row:
        """
        Get by the primary key.
        """
        if not isinstance(source_row, Row):
            if isinstance(source_row, list):
                source_row = self.Row(zip(self.primary_key, source_row))
            elif isinstance(source_row, dict):
                source_row = self.Row(source_row)
            else:
                source_row = self.Row(zip(self.primary_key, [source_row]))
        return self.get_by_lookup(
            self._get_pk_lookup_name(),
            source_row,
            stats_id=stats_id,
            parent_stats=parent_stats
        )

    def get_by_lookup(self,
                      lookup_name: str,
                      source_row: Row,
                      stats_id: str = 'get_by_lookup',
                      parent_stats: typing.Optional[Statistics] = None,
                      fallback_to_db: bool = False,
                      ) -> Row:

        if lookup_name is None:
            lookup_name = self._get_pk_lookup_name()

        fallback_to_db = fallback_to_db or self.always_fallback_to_db or not self.cache_clean

        return super().get_by_lookup(
            lookup_name=lookup_name,
            source_row=source_row,
            stats_id=stats_id,
            parent_stats=parent_stats,
            fallback_to_db=fallback_to_db,
        )



from collections.abc import MutableSet


class CaseInsentiveSet(MutableSet):
    """
    Implements a case-insensitive set while still preserving the case of the first instance of each item.
    """
    def __init__(self, values):
        self._values = {}
        for v in values:
            self.add(v)

    def __repr__(self):
        return f'<{type(self).__name__}{tuple(self._values.values())} @ {id(self):x}>'

    def __contains__(self, value: str):
        return value.casefold() in self._values

    def __iter__(self):
        return iter(self._values.values())

    def __len__(self):
        return len(self._values)

    def add(self, value: str):
        value_key = value.casefold()
        if value_key not in self._values:
            self._values[value_key] = value

    def discard(self, value: str):
        try:
            del self._values[value.casefold()]
        except KeyError:
            pass
